# Copyright 2016 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import io
import operator
import tempfile
import uuid

from openstack.cloud import exc
from openstack.cloud import meta
from openstack import exceptions
from openstack.image.v1 import image as image_v1
from openstack.image.v2 import image
from openstack.tests import fakes
from openstack.tests.unit import base


IMPORT_METHODS = 'glance-direct,web-download'


class BaseTestImage(base.TestCase):

    def setUp(self):
        super(BaseTestImage, self).setUp()
        self.image_id = str(uuid.uuid4())
        self.image_name = self.getUniqueString('image')
        self.object_name = u'images/{name}'.format(name=self.image_name)
        self.imagefile = tempfile.NamedTemporaryFile(delete=False)
        data = b'\2\0'
        self.imagefile.write(data)
        self.imagefile.close()
        self.output = data
        self.fake_image_dict = fakes.make_fake_image(
            image_id=self.image_id, image_name=self.image_name,
            data=self.imagefile.name
        )
        self.fake_search_return = {'images': [self.fake_image_dict]}
        self.container_name = self.getUniqueString('container')

    def _compare_images(self, exp, real):
        self.assertDictEqual(
            image.Image(**exp).to_dict(computed=False),
            real.to_dict(computed=False))

    def _compare_images_v1(self, exp, real):
        self.assertDictEqual(
            image_v1.Image(**exp).to_dict(computed=False),
            real.to_dict(computed=False))


class TestImage(BaseTestImage):

    def setUp(self):
        super(TestImage, self).setUp()
        self.use_glance()

    def test_config_v1(self):
        self.cloud.config.config['image_api_version'] = '1'
        # We override the scheme of the endpoint with the scheme of the service
        # because glance has a bug where it doesn't return https properly.
        self.assertEqual(
            'https://image.example.com/v1/',
            self.cloud._image_client.get_endpoint())
        self.assertEqual(
            '1', self.cloud_config.get_api_version('image'))

    def test_config_v2(self):
        self.cloud.config.config['image_api_version'] = '2'
        # We override the scheme of the endpoint with the scheme of the service
        # because glance has a bug where it doesn't return https properly.
        self.assertEqual(
            'https://image.example.com/v2/',
            self.cloud._image_client.get_endpoint())
        self.assertEqual(
            '2', self.cloud_config.get_api_version('image'))

    def test_download_image_no_output(self):
        self.assertRaises(exc.OpenStackCloudException,
                          self.cloud.download_image, self.image_name)

    def test_download_image_two_outputs(self):
        fake_fd = io.BytesIO()
        self.assertRaises(exc.OpenStackCloudException,
                          self.cloud.download_image, self.image_name,
                          output_path='fake_path', output_file=fake_fd)

    def test_download_image_no_images_found(self):
        self.register_uris([
            dict(method='GET',
                 uri='https://image.example.com/v2/images/{name}'.format(
                     name=self.image_name),
                 status_code=404),
            dict(method='GET',
                 uri='https://image.example.com/v2/images?name={name}'.format(
                     name=self.image_name),
                 json=dict(images=[])),
            dict(method='GET',
                 uri='https://image.example.com/v2/images?os_hidden=True',
                 json=dict(images=[]))])
        self.assertRaises(exc.OpenStackCloudResourceNotFound,
                          self.cloud.download_image, self.image_name,
                          output_path='fake_path')
        self.assert_calls()

    def _register_image_mocks(self):
        self.register_uris([
            dict(method='GET',
                 uri='https://image.example.com/v2/images/{name}'.format(
                     name=self.image_name),
                 status_code=404),
            dict(method='GET',
                 uri='https://image.example.com/v2/images?name={name}'.format(
                     name=self.image_name),
                 json=self.fake_search_return),
            dict(method='GET',
                 uri='https://image.example.com/v2/images/{id}/file'.format(
                     id=self.image_id),
                 content=self.output,
                 headers={
                     'Content-Type': 'application/octet-stream',
                     'Content-MD5': self.fake_image_dict['checksum']
                 })
        ])

    def test_download_image_with_fd(self):
        self._register_image_mocks()
        output_file = io.BytesIO()
        self.cloud.download_image(self.image_name, output_file=output_file)
        output_file.seek(0)
        self.assertEqual(output_file.read(), self.output)
        self.assert_calls()

    def test_download_image_with_path(self):
        self._register_image_mocks()
        output_file = tempfile.NamedTemporaryFile()
        self.cloud.download_image(
            self.image_name, output_path=output_file.name)
        output_file.seek(0)
        self.assertEqual(output_file.read(), self.output)
        self.assert_calls()

    def test_get_image_name(self, cloud=None):
        cloud = cloud or self.cloud
        self.register_uris([
            dict(method='GET',
                 uri=self.get_mock_url(
                     'image', append=['images'], base_url_append='v2'),
                 json=self.fake_search_return),
            dict(method='GET',
                 uri=self.get_mock_url(
                     'image', append=['images'], base_url_append='v2'),
                 json=self.fake_search_return),
        ])

        self.assertEqual(
            self.image_name, cloud.get_image_name(self.image_id))
        self.assertEqual(
            self.image_name, cloud.get_image_name(self.image_name))

        self.assert_calls()

    def test_get_image_by_id(self):
        self.register_uris([
            dict(method='GET',
                 uri=self.get_mock_url(
                     'image', append=['images', self.image_id],
                     base_url_append='v2'),
                 json=self.fake_image_dict)
        ])
        self._compare_images(
            self.fake_image_dict,
            self.cloud.get_image_by_id(self.image_id)
        )
        self.assert_calls()

    def test_get_image_id(self, cloud=None):
        cloud = cloud or self.cloud
        self.register_uris([
            dict(method='GET',
                 uri=self.get_mock_url(
                     'image', append=['images'], base_url_append='v2'),
                 json=self.fake_search_return),
            dict(method='GET',
                 uri=self.get_mock_url(
                     'image', append=['images'], base_url_append='v2'),
                 json=self.fake_search_return),
        ])

        self.assertEqual(
            self.image_id, cloud.get_image_id(self.image_id))
        self.assertEqual(
            self.image_id, cloud.get_image_id(self.image_name))

        self.assert_calls()

    def test_get_image_name_operator(self):
        # This should work the same as non-operator, just verifying it does.
        self.test_get_image_name(cloud=self.cloud)

    def test_get_image_id_operator(self):
        # This should work the same as the other test, just verifying it does.
        self.test_get_image_id(cloud=self.cloud)

    def test_empty_list_images(self):
        self.register_uris([
            dict(method='GET',
                 uri=self.get_mock_url(
                     'image', append=['images'], base_url_append='v2'),
                 json={'images': []})
        ])
        self.assertEqual([], self.cloud.list_images())
        self.assert_calls()

    def test_list_images(self):
        self.register_uris([
            dict(method='GET',
                 uri=self.get_mock_url(
                     'image', append=['images'], base_url_append='v2'),
                 json=self.fake_search_return)
        ])
        [self._compare_images(a, b) for a, b in zip(
            [self.fake_image_dict],
            self.cloud.list_images())]
        self.assert_calls()

    def test_list_images_show_all(self):
        self.register_uris([
            dict(method='GET',
                 uri=self.get_mock_url(
                     'image', append=['images'], base_url_append='v2',
                     qs_elements=['member_status=all']),
                 json=self.fake_search_return)
        ])
        [self._compare_images(a, b) for a, b in zip(
            [self.fake_image_dict],
            self.cloud.list_images(show_all=True))]
        self.assert_calls()

    def test_list_images_show_all_deleted(self):
        deleted_image = self.fake_image_dict.copy()
        deleted_image['status'] = 'deleted'
        self.register_uris([
            dict(method='GET',
                 uri=self.get_mock_url(
                     'image', append=['images'], base_url_append='v2',
                     qs_elements=['member_status=all']),
                 json={'images': [self.fake_image_dict, deleted_image]})
        ])
        [self._compare_images(a, b) for a, b in zip(
            [self.fake_image_dict],
            self.cloud.list_images(show_all=True))]
        self.assert_calls()

    def test_list_images_no_filter_deleted(self):
        deleted_image = self.fake_image_dict.copy()
        deleted_image['status'] = 'deleted'
        self.register_uris([
            dict(method='GET',
                 uri=self.get_mock_url(
                     'image', append=['images'], base_url_append='v2'),
                 json={'images': [self.fake_image_dict, deleted_image]})
        ])
        [self._compare_images(a, b) for a, b in zip(
            [self.fake_image_dict],
            self.cloud.list_images(filter_deleted=False))]
        self.assert_calls()

    def test_list_images_filter_deleted(self):
        deleted_image = self.fake_image_dict.copy()
        deleted_image['status'] = 'deleted'
        self.register_uris([
            dict(method='GET',
                 uri=self.get_mock_url(
                     'image', append=['images'], base_url_append='v2'),
                 json={'images': [self.fake_image_dict, deleted_image]})
        ])
        [self._compare_images(a, b) for a, b in zip(
            [self.fake_image_dict],
            self.cloud.list_images())]
        self.assert_calls()

    def test_list_images_string_properties(self):
        image_dict = self.fake_image_dict.copy()
        image_dict['properties'] = 'list,of,properties'
        self.register_uris([
            dict(method='GET',
                 uri=self.get_mock_url(
                     'image', append=['images'], base_url_append='v2'),
                 json={'images': [image_dict]}),
        ])
        images = self.cloud.list_images()
        [self._compare_images(a, b) for a, b in zip(
            [image_dict],
            images)]

        self.assertEqual(
            images[0]['properties']['properties'],
            'list,of,properties')
        self.assert_calls()

    def test_list_images_paginated(self):
        marker = str(uuid.uuid4())
        self.register_uris([
            dict(method='GET',
                 uri=self.get_mock_url(
                     'image', append=['images'], base_url_append='v2'),
                 json={'images': [self.fake_image_dict],
                       'next': '/v2/images?marker={marker}'.format(
                           marker=marker)}),
            dict(method='GET',
                 uri=self.get_mock_url(
                     'image', append=['images'], base_url_append='v2',
                     qs_elements=['marker={marker}'.format(marker=marker)]),
                 json=self.fake_search_return)
        ])
        [self._compare_images(a, b) for a, b in zip(
            [self.fake_image_dict],
            self.cloud.list_images())]
        self.assert_calls()

    def test_create_image_put_v2_no_import(self):
        self.cloud.image_api_use_tasks = False

        self.register_uris([
            dict(method='GET',
                 uri=self.get_mock_url(
                     'image', append=['images', self.image_name],
                     base_url_append='v2'),
                 status_code=404),
            dict(method='GET',
                 uri=self.get_mock_url(
                     'image', append=['images'],
                     base_url_append='v2',
                     qs_elements=['name=' + self.image_name]),
                 validate=dict(),
                 json={'images': []}),
            dict(method='GET',
                 uri=self.get_mock_url(
                     'image', append=['images'],
                     base_url_append='v2',
                     qs_elements=['os_hidden=True']),
                 json={'images': []}),
            dict(method='POST',
                 uri=self.get_mock_url(
                     'image', append=['images'], base_url_append='v2'),
                 json=self.fake_image_dict,
                 validate=dict(
                     json={
                         u'container_format': u'bare',
                         u'disk_format': u'qcow2',
                         u'name': self.image_name,
                         u'owner_specified.openstack.md5':
                             self.fake_image_dict[
                                 'owner_specified.openstack.md5'],
                         u'owner_specified.openstack.object': self.object_name,
                         u'owner_specified.openstack.sha256':
                             self.fake_image_dict[
                                 'owner_specified.openstack.sha256'],
                         u'visibility': u'private',
                         u'tags': [u'tag1', u'tag2']})
                 ),
            dict(method='PUT',
                 uri=self.get_mock_url(
                     'image', append=['images', self.image_id, 'file'],
                     base_url_append='v2'),
                 request_headers={'Content-Type': 'application/octet-stream'}),
            dict(method='GET',
                 uri=self.get_mock_url(
                     'image', append=['images', self.fake_image_dict['id']],
                     base_url_append='v2'
                 ),
                 json=self.fake_image_dict),
            dict(method='GET',
                 uri=self.get_mock_url(
                     'image', append=['images'], base_url_append='v2'),
                 complete_qs=True,
                 json=self.fake_search_return)
        ])

        self.cloud.create_image(
            self.image_name, self.imagefile.name, wait=True, timeout=1,
            tags=['tag1', 'tag2'],
            is_public=False, validate_checksum=True)

        self.assert_calls()
        self.assertEqual(self.adapter.request_history[7].text.read(),
                         self.output)

    def test_create_image_put_v2_import_supported(self):
        self.cloud.image_api_use_tasks = False

        self.register_uris([
            dict(method='GET',
                 uri=self.get_mock_url(
                     'image', append=['images', self.image_name],
                     base_url_append='v2'),
                 status_code=404),
            dict(method='GET',
                 uri=self.get_mock_url(
                     'image', append=['images'],
                     base_url_append='v2',
                     qs_elements=['name=' + self.image_name]),
                 validate=dict(),
                 json={'images': []}),
            dict(method='GET',
                 uri=self.get_mock_url(
                     'image', append=['images'],
                     base_url_append='v2',
                     qs_elements=['os_hidden=True']),
                 json={'images': []}),
            dict(method='POST',
                 uri=self.get_mock_url(
                     'image', append=['images'], base_url_append='v2'),
                 json=self.fake_image_dict,
                 headers={
                     'OpenStack-image-import-methods': IMPORT_METHODS,
                 },
                 validate=dict(
                     json={
                         u'container_format': u'bare',
                         u'disk_format': u'qcow2',
                         u'name': self.image_name,
                         u'owner_specified.openstack.md5':
                             self.fake_image_dict[
                                 'owner_specified.openstack.md5'],
                         u'owner_specified.openstack.object': self.object_name,
                         u'owner_specified.openstack.sha256':
                             self.fake_image_dict[
                                 'owner_specified.openstack.sha256'],
                         u'visibility': u'private',
                         u'tags': [u'tag1', u'tag2']})
                 ),
            dict(method='PUT',
                 uri=self.get_mock_url(
                     'image', append=['images', self.image_id, 'file'],
                     base_url_append='v2'),
                 request_headers={'Content-Type': 'application/octet-stream'}),
            dict(method='GET',
                 uri=self.get_mock_url(
                     'image', append=['images', self.fake_image_dict['id']],
                     base_url_append='v2'
                 ),
                 json=self.fake_image_dict),
            dict(method='GET',
                 uri=self.get_mock_url(
                     'image', append=['images'], base_url_append='v2'),
                 complete_qs=True,
                 json=self.fake_search_return)
        ])

        self.cloud.create_image(
            self.image_name, self.imagefile.name, wait=True, timeout=1,
            tags=['tag1', 'tag2'],
            is_public=False, validate_checksum=True)

        self.assert_calls()
        self.assertEqual(self.adapter.request_history[7].text.read(),
                         self.output)

    def test_create_image_use_import(self):
        self.cloud.image_api_use_tasks = False

        self.register_uris([
            dict(method='GET',
                 uri=self.get_mock_url(
                     'image', append=['images', self.image_name],
                     base_url_append='v2'),
                 status_code=404),
            dict(method='GET',
                 uri=self.get_mock_url(
                     'image', append=['images'],
                     base_url_append='v2',
                     qs_elements=['name=' + self.image_name]),
                 validate=dict(),
                 json={'images': []}),
            dict(method='GET',
                 uri=self.get_mock_url(
                     'image', append=['images'],
                     base_url_append='v2',
                     qs_elements=['os_hidden=True']),
                 json={'images': []}),
            dict(method='POST',
                 uri=self.get_mock_url(
                     'image', append=['images'], base_url_append='v2'),
                 json=self.fake_image_dict,
                 headers={
                     'OpenStack-image-import-methods': IMPORT_METHODS,
                 },
                 validate=dict(
                     json={
                         u'container_format': u'bare',
                         u'disk_format': u'qcow2',
                         u'name': self.image_name,
                         u'owner_specified.openstack.md5':
                             self.fake_image_dict[
                                 'owner_specified.openstack.md5'],
                         u'owner_specified.openstack.object': self.object_name,
                         u'owner_specified.openstack.sha256':
                             self.fake_image_dict[
                                 'owner_specified.openstack.sha256'],
                         u'visibility': u'private',
                         u'tags': [u'tag1', u'tag2']})
                 ),
            dict(method='PUT',
                 uri=self.get_mock_url(
                     'image', append=['images', self.image_id, 'stage'],
                     base_url_append='v2'),
                 request_headers={'Content-Type': 'application/octet-stream'}),
            dict(method='POST',
                 uri=self.get_mock_url(
                     'image', append=['images', self.image_id, 'import'],
                     base_url_append='v2'),
                 json={'method': {'name': 'glance-direct'}}),
            dict(method='GET',
                 uri=self.get_mock_url(
                     'image', append=['images', self.fake_image_dict['id']],
                     base_url_append='v2'
                 ),
                 json=self.fake_image_dict),
            dict(method='GET',
                 uri=self.get_mock_url(
                     'image', append=['images'], base_url_append='v2'),
                 complete_qs=True,
                 json=self.fake_search_return)
        ])

        self.cloud.create_image(
            self.image_name, self.imagefile.name, wait=True, timeout=1,
            tags=['tag1', 'tag2'],
            is_public=False, validate_checksum=True,
            use_import=True,
        )

        self.assert_calls()
        self.assertEqual(self.adapter.request_history[7].text.read(),
                         self.output)

    def test_create_image_task(self):
        self.cloud.image_api_use_tasks = True
        endpoint = self.cloud._object_store_client.get_endpoint()

        task_id = str(uuid.uuid4())
        args = dict(
            id=task_id,
            status='success',
            type='import',
            result={
                'image_id': self.image_id,
            },
        )

        image_no_checksums = self.fake_image_dict.copy()
        del(image_no_checksums['owner_specified.openstack.md5'])
        del(image_no_checksums['owner_specified.openstack.sha256'])
        del(image_no_checksums['owner_specified.openstack.object'])

        self.register_uris([
            dict(method='GET',
                 uri=self.get_mock_url(
                     'image', append=['images', self.image_name],
                     base_url_append='v2'),
                 status_code=404),
            dict(method='GET',
                 uri=self.get_mock_url(
                     'image', append=['images'],
                     base_url_append='v2',
                     qs_elements=['name=' + self.image_name]),
                 validate=dict(),
                 json={'images': []}),
            dict(method='GET',
                 uri=self.get_mock_url(
                     'image', append=['images'],
                     base_url_append='v2',
                     qs_elements=['os_hidden=True']),
                 json={'images': []}),
            dict(method='HEAD',
                 uri='{endpoint}/{container}'.format(
                     endpoint=endpoint, container=self.container_name),
                 status_code=404),
            dict(method='PUT',
                 uri='{endpoint}/{container}'.format(
                     endpoint=endpoint, container=self.container_name),
                 status_code=201,
                 headers={'Date': 'Fri, 16 Dec 2016 18:21:20 GMT',
                          'Content-Length': '0',
                          'Content-Type': 'text/html; charset=UTF-8'}),
            dict(method='HEAD',
                 uri='{endpoint}/{container}'.format(
                     endpoint=endpoint, container=self.container_name),
                 headers={
                     'Content-Length': '0',
                     'X-Container-Object-Count': '0',
                     'Accept-Ranges': 'bytes',
                     'X-Storage-Policy': 'Policy-0',
                     'Date': 'Fri, 16 Dec 2016 18:29:05 GMT',
                     'X-Timestamp': '1481912480.41664',
                     'X-Trans-Id': 'tx60ec128d9dbf44b9add68-0058543271dfw1',
                     'X-Container-Bytes-Used': '0',
                     'Content-Type': 'text/plain; charset=utf-8'}),
            dict(method='GET',
                 # This is explicitly not using get_mock_url because that
                 # gets us a project-id oriented URL.
                 uri='https://object-store.example.com/info',
                 json=dict(
                     swift={'max_file_size': 1000},
                     slo={'min_segment_size': 500})),
            dict(method='HEAD',
                 uri='{endpoint}/{container}/{object}'.format(
                     endpoint=endpoint, container=self.container_name,
                     object=self.image_name),
                 status_code=404),
            dict(method='PUT',
                 uri='{endpoint}/{container}/{object}'.format(
                     endpoint=endpoint, container=self.container_name,
                     object=self.image_name),
                 status_code=201,
                 validate=dict(
                     headers={'X-Object-Meta-x-sdk-md5':
                              self.fake_image_dict[
                                  'owner_specified.openstack.md5'],
                              'X-Object-Meta-x-sdk-sha256':
                              self.fake_image_dict[
                                  'owner_specified.openstack.sha256']})
                 ),
            dict(method='POST',
                 uri=self.get_mock_url(
                     'image', append=['tasks'], base_url_append='v2'),
                 json={'id': task_id, 'status': 'processing'},
                 validate=dict(
                     json=dict(
                         type='import', input={
                             'import_from': '{container}/{object}'.format(
                                 container=self.container_name,
                                 object=self.image_name),
                             'image_properties': {'name': self.image_name}}))
                 ),
            dict(method='GET',
                 uri=self.get_mock_url(
                     'image', append=['tasks', task_id], base_url_append='v2'),
                 status_code=503, text='Random error'),
            dict(method='GET',
                 uri=self.get_mock_url(
                     'image', append=['tasks', task_id], base_url_append='v2'),
                 json=args),
            dict(method='GET',
                 uri=self.get_mock_url(
                     'image', append=['images', self.image_id],
                     base_url_append='v2'),
                 json=image_no_checksums),
            dict(method='PATCH',
                 uri=self.get_mock_url(
                     'image', append=['images', self.image_id],
                     base_url_append='v2'),
                 validate=dict(
                     json=sorted([
                         {u'op': u'add',
                          u'value': '{container}/{object}'.format(
                              container=self.container_name,
                              object=self.image_name),
                          u'path': u'/owner_specified.openstack.object'},
                         {u'op': u'add', u'value':
                             self.fake_image_dict[
                                 'owner_specified.openstack.md5'],
                          u'path': u'/owner_specified.openstack.md5'},
                         {u'op': u'add', u'value':
                             self.fake_image_dict[
                                 'owner_specified.openstack.sha256'],
                          u'path': u'/owner_specified.openstack.sha256'}],
                         key=operator.itemgetter('path')),
                     headers={
                         'Content-Type':
                             'application/openstack-images-v2.1-json-patch'}),
                 json=self.fake_search_return
                 ),
            dict(method='HEAD',
                 uri='{endpoint}/{container}/{object}'.format(
                     endpoint=endpoint, container=self.container_name,
                     object=self.image_name),
                 headers={
                     'X-Timestamp': '1429036140.50253',
                     'X-Trans-Id': 'txbbb825960a3243b49a36f-005a0dadaedfw1',
                     'Content-Length': '1290170880',
                     'Last-Modified': 'Tue, 14 Apr 2015 18:29:01 GMT',
                     'X-Object-Meta-X-Sdk-Sha256':
                     self.fake_image_dict[
                         'owner_specified.openstack.sha256'],
                     'X-Object-Meta-X-Sdk-Md5':
                     self.fake_image_dict[
                         'owner_specified.openstack.md5'],
                     'Date': 'Thu, 16 Nov 2017 15:24:30 GMT',
                     'Accept-Ranges': 'bytes',
                     'Content-Type': 'application/octet-stream',
                     'Etag': fakes.NO_MD5}),
            dict(method='DELETE',
                 uri='{endpoint}/{container}/{object}'.format(
                     endpoint=endpoint, container=self.container_name,
                     object=self.image_name)),
            dict(method='GET',
                 uri=self.get_mock_url(
                     'image', append=['images'], base_url_append='v2'),
                 complete_qs=True,
                 json=self.fake_search_return)
        ])

        self.cloud.create_image(
            self.image_name, self.imagefile.name, wait=True, timeout=1,
            disk_format='vhd', container_format='ovf',
            is_public=False, validate_checksum=True,
            container=self.container_name)

        self.assert_calls()

    def test_delete_autocreated_no_tasks(self):
        self.use_keystone_v3()
        self.cloud.image_api_use_tasks = False
        deleted = self.cloud.delete_autocreated_image_objects(
            container=self.container_name)
        self.assertFalse(deleted)
        self.assert_calls([])

    def test_delete_image_task(self):
        self.cloud.image_api_use_tasks = True
        endpoint = self.cloud._object_store_client.get_endpoint()

        object_path = self.fake_image_dict['owner_specified.openstack.object']

        image_no_checksums = self.fake_image_dict.copy()
        del(image_no_checksums['owner_specified.openstack.md5'])
        del(image_no_checksums['owner_specified.openstack.sha256'])
        del(image_no_checksums['owner_specified.openstack.object'])

        self.register_uris([
            dict(method='GET',
                 uri=self.get_mock_url(
                     'image', append=['images'], base_url_append='v2'),
                 json=self.fake_search_return),
            dict(method='DELETE',
                 uri='https://image.example.com/v2/images/{id}'.format(
                     id=self.image_id)),
            dict(method='HEAD',
                 uri='{endpoint}/{object}'.format(
                     endpoint=endpoint,
                     object=object_path),
                 headers={
                     'X-Timestamp': '1429036140.50253',
                     'X-Trans-Id': 'txbbb825960a3243b49a36f-005a0dadaedfw1',
                     'Content-Length': '1290170880',
                     'Last-Modified': 'Tue, 14 Apr 2015 18:29:01 GMT',
                     'X-Object-Meta-X-Sdk-Sha256':
                     self.fake_image_dict[
                         'owner_specified.openstack.sha256'],
                     'X-Object-Meta-X-Sdk-Md5':
                     self.fake_image_dict[
                         'owner_specified.openstack.md5'],
                     'Date': 'Thu, 16 Nov 2017 15:24:30 GMT',
                     'Accept-Ranges': 'bytes',
                     'Content-Type': 'application/octet-stream',
                     'Etag': fakes.NO_MD5}),
            dict(method='DELETE',
                 uri='{endpoint}/{object}'.format(
                     endpoint=endpoint,
                     object=object_path)),
        ])

        self.cloud.delete_image(self.image_id)

        self.assert_calls()

    def test_delete_autocreated_image_objects(self):
        self.use_keystone_v3()
        self.cloud.image_api_use_tasks = True
        endpoint = self.cloud._object_store_client.get_endpoint()
        other_image = self.getUniqueString('no-delete')

        self.register_uris([
            dict(method='GET',
                 uri=self.get_mock_url(
                     service_type='object-store',
                     resource=self.container_name,
                     qs_elements=['format=json']),
                 json=[{
                     'content_type': 'application/octet-stream',
                     'bytes': 1437258240,
                     'hash': '249219347276c331b87bf1ac2152d9af',
                     'last_modified': '2015-02-16T17:50:05.289600',
                     'name': other_image,
                 }, {
                     'content_type': 'application/octet-stream',
                     'bytes': 1290170880,
                     'hash': fakes.NO_MD5,
                     'last_modified': '2015-04-14T18:29:00.502530',
                     'name': self.image_name,
                 }]),
            dict(method='HEAD',
                 uri=self.get_mock_url(
                     service_type='object-store',
                     resource=self.container_name,
                     append=[other_image]),
                 headers={
                     'X-Timestamp': '1429036140.50253',
                     'X-Trans-Id': 'txbbb825960a3243b49a36f-005a0dadaedfw1',
                     'Content-Length': '1290170880',
                     'Last-Modified': 'Tue, 14 Apr 2015 18:29:01 GMT',
                     'X-Object-Meta-X-Shade-Sha256': 'does not matter',
                     'X-Object-Meta-X-Shade-Md5': 'does not matter',
                     'Date': 'Thu, 16 Nov 2017 15:24:30 GMT',
                     'Accept-Ranges': 'bytes',
                     'Content-Type': 'application/octet-stream',
                     'Etag': '249219347276c331b87bf1ac2152d9af',
                 }),
            dict(method='HEAD',
                 uri=self.get_mock_url(
                     service_type='object-store',
                     resource=self.container_name,
                     append=[self.image_name]),
                 headers={
                     'X-Timestamp': '1429036140.50253',
                     'X-Trans-Id': 'txbbb825960a3243b49a36f-005a0dadaedfw1',
                     'Content-Length': '1290170880',
                     'Last-Modified': 'Tue, 14 Apr 2015 18:29:01 GMT',
                     'X-Object-Meta-X-Shade-Sha256': fakes.NO_SHA256,
                     'X-Object-Meta-X-Shade-Md5': fakes.NO_MD5,
                     'Date': 'Thu, 16 Nov 2017 15:24:30 GMT',
                     'Accept-Ranges': 'bytes',
                     'Content-Type': 'application/octet-stream',
                     ('X-Object-Meta-'
                      + self.cloud._OBJECT_AUTOCREATE_KEY): 'true',
                     'Etag': fakes.NO_MD5,
                     'X-Static-Large-Object': 'false'}),
            dict(method='DELETE',
                 uri='{endpoint}/{container}/{object}'.format(
                     endpoint=endpoint, container=self.container_name,
                     object=self.image_name)),
        ])

        deleted = self.cloud.delete_autocreated_image_objects(
            container=self.container_name)
        self.assertTrue(deleted)

        self.assert_calls()

    def _image_dict(self, fake_image):
        return self.cloud._normalize_image(meta.obj_to_munch(fake_image))

    def _call_create_image(self, name, **kwargs):
        imagefile = tempfile.NamedTemporaryFile(delete=False)
        imagefile.write(b'\0')
        imagefile.close()
        self.cloud.create_image(
            name, imagefile.name, wait=True, timeout=1,
            is_public=False, validate_checksum=True, **kwargs)

    def test_create_image_put_v1(self):
        self.cloud.config.config['image_api_version'] = '1'

        args = {'name': self.image_name,
                'container_format': 'bare', 'disk_format': 'qcow2',
                'properties': {
                    'owner_specified.openstack.md5': fakes.NO_MD5,
                    'owner_specified.openstack.sha256': fakes.NO_SHA256,
                    'owner_specified.openstack.object': 'images/{name}'.format(
                        name=self.image_name),
                    'is_public': False}}

        ret = args.copy()
        ret['id'] = self.image_id
        ret['status'] = 'success'

        self.register_uris([
            dict(method='GET',
                 uri='https://image.example.com/v1/images/' + self.image_name,
                 status_code=404),
            dict(method='GET',
                 uri='https://image.example.com/v1/images/detail?name='
                 + self.image_name,
                 json={'images': []}),
            dict(method='POST',
                 uri='https://image.example.com/v1/images',
                 json={'image': ret},
                 validate=dict(json=args)),
            dict(method='PUT',
                 uri='https://image.example.com/v1/images/{id}'.format(
                     id=self.image_id),
                 json={'image': ret},
                 validate=dict(headers={
                     'x-image-meta-checksum': fakes.NO_MD5,
                     'x-glance-registry-purge-props': 'false'
                 })),
            dict(method='GET',
                 uri='https://image.example.com/v1/images/detail',
                 json={'images': [ret]}),
        ])
        self._call_create_image(self.image_name)
        [self._compare_images_v1(b, a) for a, b in zip(
            self.cloud.list_images(), [ret])]

    def test_create_image_put_v1_bad_delete(self):
        self.cloud.config.config['image_api_version'] = '1'

        args = {'name': self.image_name,
                'container_format': 'bare', 'disk_format': 'qcow2',
                'properties': {
                    'owner_specified.openstack.md5': fakes.NO_MD5,
                    'owner_specified.openstack.sha256': fakes.NO_SHA256,
                    'owner_specified.openstack.object': 'images/{name}'.format(
                        name=self.image_name),
                    'is_public': False},
                'validate_checksum': True}

        ret = args.copy()
        ret['id'] = self.image_id
        ret['status'] = 'success'

        self.register_uris([
            dict(method='GET',
                 uri='https://image.example.com/v1/images/' + self.image_name,
                 status_code=404),
            dict(method='GET',
                 uri='https://image.example.com/v1/images/detail?name='
                 + self.image_name,
                 json={'images': []}),
            dict(method='POST',
                 uri='https://image.example.com/v1/images',
                 json={'image': ret},
                 validate=dict(json=args)),
            dict(method='PUT',
                 uri='https://image.example.com/v1/images/{id}'.format(
                     id=self.image_id),
                 status_code=400,
                 validate=dict(headers={
                     'x-image-meta-checksum': fakes.NO_MD5,
                     'x-glance-registry-purge-props': 'false'
                 })),
            dict(method='DELETE',
                 uri='https://image.example.com/v1/images/{id}'.format(
                     id=self.image_id),
                 json={'images': [ret]}),
        ])

        self.assertRaises(
            exc.OpenStackCloudHTTPError,
            self._call_create_image,
            self.image_name)

        self.assert_calls()

    def test_update_image_no_patch(self):
        self.cloud.image_api_use_tasks = False

        args = {'name': self.image_name,
                'container_format': 'bare', 'disk_format': 'qcow2',
                'owner_specified.openstack.md5': fakes.NO_MD5,
                'owner_specified.openstack.sha256': fakes.NO_SHA256,
                'owner_specified.openstack.object': 'images/{name}'.format(
                    name=self.image_name),
                'visibility': 'private'}

        ret = args.copy()
        ret['id'] = self.image_id
        ret['status'] = 'success'

        self.cloud.update_image_properties(
            image=self._image_dict(ret),
            **{'owner_specified.openstack.object': 'images/{name}'.format(
                name=self.image_name)})

        self.assert_calls()

    def test_create_image_put_v2_bad_delete(self):
        self.cloud.image_api_use_tasks = False

        args = {'name': self.image_name,
                'container_format': 'bare', 'disk_format': 'qcow2',
                'owner_specified.openstack.md5': fakes.NO_MD5,
                'owner_specified.openstack.sha256': fakes.NO_SHA256,
                'owner_specified.openstack.object': 'images/{name}'.format(
                    name=self.image_name),
                'visibility': 'private'}

        ret = args.copy()
        ret['id'] = self.image_id
        ret['status'] = 'success'

        self.register_uris([
            dict(method='GET',
                 uri=self.get_mock_url(
                     'image', append=['images', self.image_name],
                     base_url_append='v2'),
                 status_code=404),
            dict(method='GET',
                 uri=self.get_mock_url(
                     'image', append=['images'],
                     base_url_append='v2',
                     qs_elements=['name=' + self.image_name]),
                 validate=dict(),
                 json={'images': []}),
            dict(method='GET',
                 uri=self.get_mock_url(
                     'image', append=['images'],
                     base_url_append='v2',
                     qs_elements=['os_hidden=True']),
                 json={'images': []}),
            dict(method='POST',
                 uri='https://image.example.com/v2/images',
                 json=ret,
                 validate=dict(json=args)),
            dict(method='PUT',
                 uri='https://image.example.com/v2/images/{id}/file'.format(
                     id=self.image_id),
                 status_code=400,
                 validate=dict(
                     headers={
                         'Content-Type': 'application/octet-stream',
                     },
                 )),
            dict(method='DELETE',
                 uri='https://image.example.com/v2/images/{id}'.format(
                     id=self.image_id)),
        ])

        self.assertRaises(
            exc.OpenStackCloudHTTPError,
            self._call_create_image,
            self.image_name)

        self.assert_calls()

    def test_create_image_put_v2_wrong_checksum_delete(self):
        self.cloud.image_api_use_tasks = False

        fake_image = self.fake_image_dict

        fake_image['owner_specified.openstack.md5'] = 'a'
        fake_image['owner_specified.openstack.sha256'] = 'b'

        self.register_uris([
            dict(method='POST',
                 uri=self.get_mock_url(
                     'image', append=['images'], base_url_append='v2'),
                 json=self.fake_image_dict,
                 validate=dict(
                     json={
                         u'container_format': u'bare',
                         u'disk_format': u'qcow2',
                         u'name': self.image_name,
                         u'owner_specified.openstack.md5':
                             fake_image[
                                 'owner_specified.openstack.md5'],
                         u'owner_specified.openstack.object': self.object_name,
                         u'owner_specified.openstack.sha256':
                             fake_image[
                                 'owner_specified.openstack.sha256'],
                         u'visibility': u'private'})
                 ),
            dict(method='PUT',
                 uri=self.get_mock_url(
                     'image', append=['images', self.image_id, 'file'],
                     base_url_append='v2'),
                 request_headers={'Content-Type': 'application/octet-stream'}),
            dict(method='GET',
                 uri=self.get_mock_url(
                     'image', append=['images', self.fake_image_dict['id']],
                     base_url_append='v2'
                 ),
                 json=fake_image),
            dict(method='DELETE',
                 uri='https://image.example.com/v2/images/{id}'.format(
                     id=self.image_id))
        ])

        self.assertRaises(
            exceptions.SDKException,
            self.cloud.create_image,
            self.image_name, self.imagefile.name,
            is_public=False, md5='a', sha256='b',
            allow_duplicates=True, validate_checksum=True
        )

        self.assert_calls()

    def test_create_image_put_bad_int(self):
        self.cloud.image_api_use_tasks = False

        self.assertRaises(
            exc.OpenStackCloudException,
            self._call_create_image, self.image_name,
            allow_duplicates=True,
            min_disk='fish', min_ram=0)

        self.assert_calls()

    def test_create_image_put_user_int(self):
        self.cloud.image_api_use_tasks = False

        args = {'name': self.image_name,
                'container_format': 'bare', 'disk_format': u'qcow2',
                'owner_specified.openstack.md5': fakes.NO_MD5,
                'owner_specified.openstack.sha256': fakes.NO_SHA256,
                'owner_specified.openstack.object': 'images/{name}'.format(
                    name=self.image_name),
                'int_v': '12345',
                'visibility': 'private',
                'min_disk': 0, 'min_ram': 0}

        ret = args.copy()
        ret['id'] = self.image_id
        ret['status'] = 'success'

        self.register_uris([
            dict(method='GET',
                 uri=self.get_mock_url(
                     'image', append=['images', self.image_name],
                     base_url_append='v2'),
                 status_code=404),
            dict(method='GET',
                 uri=self.get_mock_url(
                     'image', append=['images'],
                     base_url_append='v2',
                     qs_elements=['name=' + self.image_name]),
                 validate=dict(),
                 json={'images': []}),
            dict(method='GET',
                 uri=self.get_mock_url(
                     'image', append=['images'],
                     base_url_append='v2',
                     qs_elements=['os_hidden=True']),
                 json={'images': []}),
            dict(method='POST',
                 uri='https://image.example.com/v2/images',
                 json=ret,
                 validate=dict(json=args)),
            dict(method='PUT',
                 uri='https://image.example.com/v2/images/{id}/file'.format(
                     id=self.image_id),
                 validate=dict(
                     headers={
                         'Content-Type': 'application/octet-stream',
                     },
                 )),
            dict(method='GET',
                 uri='https://image.example.com/v2/images/{id}'.format(
                     id=self.image_id
                 ),
                 json=ret),
            dict(method='GET',
                 uri='https://image.example.com/v2/images',
                 complete_qs=True,
                 json={'images': [ret]}),
        ])

        self._call_create_image(
            self.image_name, min_disk='0', min_ram=0, int_v=12345)

        self.assert_calls()

    def test_create_image_put_meta_int(self):
        self.cloud.image_api_use_tasks = False

        args = {'name': self.image_name,
                'container_format': 'bare', 'disk_format': u'qcow2',
                'owner_specified.openstack.md5': fakes.NO_MD5,
                'owner_specified.openstack.sha256': fakes.NO_SHA256,
                'owner_specified.openstack.object': 'images/{name}'.format(
                    name=self.image_name),
                'int_v': 12345,
                'visibility': 'private',
                'min_disk': 0, 'min_ram': 0}

        ret = args.copy()
        ret['id'] = self.image_id
        ret['status'] = 'success'
        ret['checksum'] = fakes.NO_MD5

        self.register_uris([
            dict(method='GET',
                 uri=self.get_mock_url(
                     'image', append=['images', self.image_name],
                     base_url_append='v2'),
                 status_code=404),
            dict(method='GET',
                 uri=self.get_mock_url(
                     'image', append=['images'],
                     base_url_append='v2',
                     qs_elements=['name=' + self.image_name]),
                 validate=dict(),
                 json={'images': []}),
            dict(method='GET',
                 uri=self.get_mock_url(
                     'image', append=['images'],
                     base_url_append='v2',
                     qs_elements=['os_hidden=True']),
                 json={'images': []}),
            dict(method='POST',
                 uri='https://image.example.com/v2/images',
                 json=ret,
                 validate=dict(json=args)),
            dict(method='PUT',
                 uri='https://image.example.com/v2/images/{id}/file'.format(
                     id=self.image_id),
                 validate=dict(
                     headers={
                         'Content-Type': 'application/octet-stream',
                     },
                 )),
            dict(method='GET',
                 uri='https://image.example.com/v2/images/{id}'.format(
                     id=self.image_id
                 ),
                 json=ret),
            dict(method='GET',
                 uri='https://image.example.com/v2/images',
                 complete_qs=True,
                 json={'images': [ret]}),
        ])

        self._call_create_image(
            self.image_name, min_disk='0', min_ram=0, meta={'int_v': 12345})

        self.assert_calls()

    def test_create_image_put_protected(self):
        self.cloud.image_api_use_tasks = False

        args = {'name': self.image_name,
                'container_format': 'bare', 'disk_format': u'qcow2',
                'owner_specified.openstack.md5': fakes.NO_MD5,
                'owner_specified.openstack.sha256': fakes.NO_SHA256,
                'owner_specified.openstack.object': 'images/{name}'.format(
                    name=self.image_name),
                'int_v': '12345',
                'protected': False,
                'visibility': 'private',
                'min_disk': 0, 'min_ram': 0}

        ret = args.copy()
        ret['id'] = self.image_id
        ret['status'] = 'success'
        ret['checksum'] = fakes.NO_MD5

        self.register_uris([
            dict(method='GET',
                 uri=self.get_mock_url(
                     'image', append=['images', self.image_name],
                     base_url_append='v2'),
                 status_code=404),
            dict(method='GET',
                 uri=self.get_mock_url(
                     'image', append=['images'],
                     base_url_append='v2',
                     qs_elements=['name=' + self.image_name]),
                 validate=dict(),
                 json={'images': []}),
            dict(method='GET',
                 uri=self.get_mock_url(
                     'image', append=['images'],
                     base_url_append='v2',
                     qs_elements=['os_hidden=True']),
                 json={'images': []}),
            dict(method='POST',
                 uri='https://image.example.com/v2/images',
                 json=ret,
                 validate=dict(json=args)),
            dict(method='PUT',
                 uri='https://image.example.com/v2/images/{id}/file'.format(
                     id=self.image_id),
                 validate=dict(
                     headers={
                         'Content-Type': 'application/octet-stream',
                     },
                 )),
            dict(method='GET',
                 uri='https://image.example.com/v2/images/{id}'.format(
                     id=self.image_id
                 ),
                 json=ret),
            dict(method='GET',
                 uri='https://image.example.com/v2/images',
                 complete_qs=True,
                 json={'images': [ret]}),
        ])

        self._call_create_image(
            self.image_name, min_disk='0', min_ram=0,
            properties={'int_v': 12345}, is_protected=False)

        self.assert_calls()


class TestImageSuburl(BaseTestImage):

    def setUp(self):
        super(TestImageSuburl, self).setUp()
        self.os_fixture.use_suburl()
        self.os_fixture.build_tokens()
        self.use_keystone_v3()
        self.use_glance(
            image_version_json='image-version-suburl.json',
            image_discovery_url='https://example.com/image')

    def test_list_images(self):
        self.register_uris([
            dict(method='GET',
                 uri=self.get_mock_url(
                     'image', append=['images'], base_url_append='v2'),
                 json=self.fake_search_return)
        ])
        [self._compare_images(b, a) for a, b in zip(
            self.cloud.list_images(),
            [self.fake_image_dict])]
        self.assert_calls()

    def test_list_images_paginated(self):
        marker = str(uuid.uuid4())
        self.register_uris([
            dict(method='GET',
                 uri=self.get_mock_url(
                     'image', append=['images'], base_url_append='v2'),
                 json={'images': [self.fake_image_dict],
                       'next': '/v2/images?marker={marker}'.format(
                           marker=marker)}),
            dict(method='GET',
                 uri=self.get_mock_url(
                     'image', append=['images'], base_url_append='v2',
                     qs_elements=['marker={marker}'.format(marker=marker)]),
                 json=self.fake_search_return)
        ])
        [self._compare_images(b, a) for a, b in zip(
            self.cloud.list_images(),
            [self.fake_image_dict, self.fake_image_dict])]
        self.assert_calls()


class TestImageV1Only(base.TestCase):

    def setUp(self):
        super(TestImageV1Only, self).setUp()
        self.use_glance(image_version_json='image-version-v1.json')

    def test_config_v1(self):

        self.cloud.config.config['image_api_version'] = '1'
        # We override the scheme of the endpoint with the scheme of the service
        # because glance has a bug where it doesn't return https properly.
        self.assertEqual(
            'https://image.example.com/v1/',
            self.cloud._image_client.get_endpoint())
        self.assertTrue(self.cloud._is_client_version('image', 1))

    def test_config_v2(self):
        self.cloud.config.config['image_api_version'] = '2'
        # We override the scheme of the endpoint with the scheme of the service
        # because glance has a bug where it doesn't return https properly.
        self.assertEqual(
            'https://image.example.com/v1/',
            self.cloud._image_client.get_endpoint())
        self.assertFalse(self.cloud._is_client_version('image', 2))


class TestImageV2Only(base.TestCase):

    def setUp(self):
        super(TestImageV2Only, self).setUp()
        self.use_glance(image_version_json='image-version-v2.json')

    def test_config_v1(self):
        self.cloud.config.config['image_api_version'] = '1'
        # We override the scheme of the endpoint with the scheme of the service
        # because glance has a bug where it doesn't return https properly.
        self.assertEqual(
            'https://image.example.com/v2/',
            self.cloud._image_client.get_endpoint())
        self.assertTrue(self.cloud._is_client_version('image', 2))

    def test_config_v2(self):
        self.cloud.config.config['image_api_version'] = '2'
        # We override the scheme of the endpoint with the scheme of the service
        # because glance has a bug where it doesn't return https properly.
        self.assertEqual(
            'https://image.example.com/v2/',
            self.cloud._image_client.get_endpoint())
        self.assertTrue(self.cloud._is_client_version('image', 2))


class TestImageVolume(BaseTestImage):

    def setUp(self):
        super(TestImageVolume, self).setUp()
        self.volume_id = str(uuid.uuid4())

    def test_create_image_volume(self):
        self.register_uris([
            self.get_cinder_discovery_mock_dict(),
            dict(method='POST',
                 uri=self.get_mock_url(
                     'volumev3', append=['volumes', self.volume_id, 'action']),
                 json={'os-volume_upload_image': {'image_id': self.image_id}},
                 validate=dict(json={
                     u'os-volume_upload_image': {
                         u'container_format': u'bare',
                         u'disk_format': u'qcow2',
                         u'force': False,
                         u'image_name': u'fake_image'}})
                 ),
            # NOTE(notmorgan): Glance discovery happens here, insert the
            # glance discovery mock at this point, DO NOT use the
            # .use_glance() method, that is intended only for use in
            # .setUp
            self.get_glance_discovery_mock_dict(),
            dict(method='GET',
                 uri=self.get_mock_url(
                     'image', append=['images'], base_url_append='v2'),
                 json=self.fake_search_return)
        ])

        self.cloud.create_image(
            'fake_image', self.imagefile.name, wait=True, timeout=1,
            volume={'id': self.volume_id})

        self.assert_calls()

    def test_create_image_volume_duplicate(self):

        self.register_uris([
            self.get_cinder_discovery_mock_dict(),
            dict(method='POST',
                 uri=self.get_mock_url(
                     'volumev3', append=['volumes', self.volume_id, 'action']),
                 json={'os-volume_upload_image': {'image_id': self.image_id}},
                 validate=dict(json={
                     u'os-volume_upload_image': {
                         u'container_format': u'bare',
                         u'disk_format': u'qcow2',
                         u'force': True,
                         u'image_name': u'fake_image'}})
                 ),
            # NOTE(notmorgan): Glance discovery happens here, insert the
            # glance discovery mock at this point, DO NOT use the
            # .use_glance() method, that is intended only for use in
            # .setUp
            self.get_glance_discovery_mock_dict(),
            dict(method='GET',
                 uri=self.get_mock_url(
                     'image', append=['images'], base_url_append='v2'),
                 json=self.fake_search_return)
        ])

        self.cloud.create_image(
            'fake_image', self.imagefile.name, wait=True, timeout=1,
            volume={'id': self.volume_id}, allow_duplicates=True)

        self.assert_calls()


class TestImageBrokenDiscovery(base.TestCase):

    def setUp(self):
        super(TestImageBrokenDiscovery, self).setUp()
        self.use_glance(image_version_json='image-version-broken.json')

    def test_url_fix(self):
        # image-version-broken.json has both http urls and localhost as the
        # host. This is testing that what is discovered is https, because
        # that's what's in the catalog, and image.example.com for the same
        # reason.
        self.register_uris([
            dict(method='GET',
                 uri=self.get_mock_url(
                     'image', append=['images'], base_url_append='v2'),
                 json={'images': []})
        ])
        self.assertEqual([], self.cloud.list_images())
        self.assertEqual(
            self.cloud._image_client.get_endpoint(),
            'https://image.example.com/v2/')
        self.assert_calls()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.cloud import exc
from openstack.tests.unit import base


class TestDomainParams(base.TestCase):

    def test_identity_params_v3(self):
        project_data = self._get_project_data(v3=True)
        self.register_uris([
            dict(method='GET',
                 uri='https://identity.example.com/v3/projects',
                 json=dict(projects=[project_data.json_response['project']]))
        ])

        ret = self.cloud._get_identity_params(
            domain_id='5678', project=project_data.project_name)
        self.assertIn('default_project_id', ret)
        self.assertEqual(ret['default_project_id'], project_data.project_id)
        self.assertIn('domain_id', ret)
        self.assertEqual(ret['domain_id'], '5678')

        self.assert_calls()

    def test_identity_params_v3_no_domain(self):
        project_data = self._get_project_data(v3=True)

        self.assertRaises(
            exc.OpenStackCloudException,
            self.cloud._get_identity_params,
            domain_id=None, project=project_data.project_name)

        self.assert_calls()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.shared_file_system.v2 import limit
from openstack.tests.unit import base

EXAMPLE = {
    "totalShareNetworksUsed": 0,
    "maxTotalShareGigabytes": 1000,
    "maxTotalShareNetworks": 10,
    "totalSharesUsed": 0,
    "totalShareGigabytesUsed": 0,
    "totalShareSnapshotsUsed": 0,
    "maxTotalShares": 50,
    "totalSnapshotGigabytesUsed": 0,
    "maxTotalSnapshotGigabytes": 1000,
    "maxTotalShareSnapshots": 50,
    "maxTotalShareReplicas": 100,
    "maxTotalReplicaGigabytes": 1000,
    "totalShareReplicasUsed": 0,
    "totalReplicaGigabytesUsed": 0
}


class TestLimit(base.TestCase):

    def test_basic(self):
        limits = limit.Limit()
        self.assertEqual('limits', limits.resources_key)
        self.assertEqual('/limits', limits.base_path)
        self.assertTrue(limits.allow_list)
        self.assertFalse(limits.allow_fetch)
        self.assertFalse(limits.allow_create)
        self.assertFalse(limits.allow_commit)
        self.assertFalse(limits.allow_delete)
        self.assertFalse(limits.allow_head)

    def test_make_limits(self):
        limits = limit.Limit(**EXAMPLE)
        self.assertEqual(EXAMPLE['totalShareNetworksUsed'],
                         limits.totalShareNetworksUsed)
        self.assertEqual(EXAMPLE['maxTotalShareGigabytes'],
                         limits.maxTotalShareGigabytes)
        self.assertEqual(EXAMPLE['maxTotalShareNetworks'],
                         limits.maxTotalShareNetworks)
        self.assertEqual(EXAMPLE['totalSharesUsed'],
                         limits.totalSharesUsed)
        self.assertEqual(EXAMPLE['totalShareGigabytesUsed'],
                         limits.totalShareGigabytesUsed)
        self.assertEqual(EXAMPLE['totalShareSnapshotsUsed'],
                         limits.totalShareSnapshotsUsed)
        self.assertEqual(EXAMPLE['maxTotalShares'],
                         limits.maxTotalShares)
        self.assertEqual(EXAMPLE['totalSnapshotGigabytesUsed'],
                         limits.totalSnapshotGigabytesUsed)
        self.assertEqual(EXAMPLE['maxTotalSnapshotGigabytes'],
                         limits.maxTotalSnapshotGigabytes)
        self.assertEqual(EXAMPLE['maxTotalShareSnapshots'],
                         limits.maxTotalShareSnapshots)
        self.assertEqual(EXAMPLE['maxTotalShareReplicas'],
                         limits.maxTotalShareReplicas)
        self.assertEqual(EXAMPLE['maxTotalReplicaGigabytes'],
                         limits.maxTotalReplicaGigabytes)
        self.assertEqual(EXAMPLE['totalShareReplicasUsed'],
                         limits.totalShareReplicasUsed)
        self.assertEqual(EXAMPLE['totalReplicaGigabytesUsed'],
                         limits.totalReplicaGigabytesUsed)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.key_manager.v1 import order
from openstack.tests.unit import base


ID_VAL = "123"
SECRET_ID = "5"
IDENTIFIER = 'http://localhost/orders/%s' % ID_VAL
EXAMPLE = {
    'created': '1',
    'creator_id': '2',
    'meta': {'key': '3'},
    'order_ref': IDENTIFIER,
    'secret_ref': 'http://localhost/secrets/%s' % SECRET_ID,
    'status': '6',
    'sub_status': '7',
    'sub_status_message': '8',
    'type': '9',
    'updated': '10'
}


class TestOrder(base.TestCase):

    def test_basic(self):
        sot = order.Order()
        self.assertIsNone(sot.resource_key)
        self.assertEqual('orders', sot.resources_key)
        self.assertEqual('/orders', sot.base_path)
        self.assertTrue(sot.allow_create)
        self.assertTrue(sot.allow_fetch)
        self.assertTrue(sot.allow_commit)
        self.assertTrue(sot.allow_delete)
        self.assertTrue(sot.allow_list)

    def test_make_it(self):
        sot = order.Order(**EXAMPLE)
        self.assertEqual(EXAMPLE['created'], sot.created_at)
        self.assertEqual(EXAMPLE['creator_id'], sot.creator_id)
        self.assertEqual(EXAMPLE['meta'], sot.meta)
        self.assertEqual(EXAMPLE['order_ref'], sot.order_ref)
        self.assertEqual(ID_VAL, sot.order_id)
        self.assertEqual(EXAMPLE['secret_ref'], sot.secret_ref)
        self.assertEqual(SECRET_ID, sot.secret_id)
        self.assertEqual(EXAMPLE['status'], sot.status)
        self.assertEqual(EXAMPLE['sub_status'], sot.sub_status)
        self.assertEqual(EXAMPLE['sub_status_message'], sot.sub_status_message)
        self.assertEqual(EXAMPLE['type'], sot.type)
        self.assertEqual(EXAMPLE['updated'], sot.updated_at)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack import resource


class PoolMember(resource.Resource):
    resource_key = 'member'
    resources_key = 'members'
    base_path = '/lbaas/pools/%(pool_id)s/members'

    _allow_unknown_attrs_in_body = True

    # capabilities
    allow_create = True
    allow_fetch = True
    allow_commit = True
    allow_delete = True
    allow_list = True

    _query_mapping = resource.QueryParameters(
        'address', 'name', 'protocol_port', 'subnet_id', 'weight',
        'project_id',
        is_admin_state_up='admin_state_up',
    )

    # Properties
    #: The ID of the owning pool
    pool_id = resource.URI('pool_id')
    #: The IP address of the pool member.
    address = resource.Body('address')
    #: The administrative state of the pool member, which is up ``True`` or
    #: down ``False``. *Type: bool*
    is_admin_state_up = resource.Body('admin_state_up', type=bool)
    #: Name of the pool member.
    name = resource.Body('name')
    #: The ID of the project this pool member is associated with.
    project_id = resource.Body('project_id', alias='tenant_id')
    #: Tenant_id (deprecated attribute).
    tenant_id = resource.Body('tenant_id', deprecated=True)
    #: The port on which the application is hosted.
    protocol_port = resource.Body('protocol_port', type=int)
    #: Subnet ID in which to access this pool member.
    subnet_id = resource.Body('subnet_id')
    #: A positive integer value that indicates the relative portion of traffic
    #: that this member should receive from the pool. For example, a member
    #: with a weight of 10 receives five times as much traffic as a member
    #: with weight of 2.
    weight = resource.Body('weight', type=int)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
test_zone
----------------------------------

Functional tests for `shade` zone methods.
"""

from testtools import content

from openstack.tests.functional import base


class TestZone(base.BaseFunctionalTest):

    def setUp(self):
        super(TestZone, self).setUp()
        if not self.user_cloud.has_service('dns'):
            self.skipTest('dns service not supported by cloud')

    def test_zones(self):
        '''Test DNS zones functionality'''
        name = 'example.net.'
        zone_type = 'primary'
        email = 'test@example.net'
        description = 'Test zone'
        ttl = 3600
        masters = None

        self.addDetail('zone', content.text_content(name))
        self.addCleanup(self.cleanup, name)

        # Test we can create a zone and we get it returned
        zone = self.user_cloud.create_zone(
            name=name, zone_type=zone_type, email=email,
            description=description, ttl=ttl,
            masters=masters)
        self.assertEqual(zone['name'], name)
        self.assertEqual(zone['type'], zone_type.upper())
        self.assertEqual(zone['email'], email)
        self.assertEqual(zone['description'], description)
        self.assertEqual(zone['ttl'], ttl)
        self.assertEqual(zone['masters'], [])

        # Test that we can list zones
        zones = self.user_cloud.list_zones()
        self.assertIsNotNone(zones)

        # Test we get the same zone with the get_zone method
        zone_get = self.user_cloud.get_zone(zone['id'])
        self.assertEqual(zone_get['id'], zone['id'])

        # Test the get method also works by name
        zone_get = self.user_cloud.get_zone(name)
        self.assertEqual(zone_get['name'], zone['name'])

        # Test we can update a field on the zone and only that field
        # is updated
        zone_update = self.user_cloud.update_zone(zone['id'], ttl=7200)
        self.assertEqual(zone_update['id'], zone['id'])
        self.assertEqual(zone_update['name'], zone['name'])
        self.assertEqual(zone_update['type'], zone['type'])
        self.assertEqual(zone_update['email'], zone['email'])
        self.assertEqual(zone_update['description'], zone['description'])
        self.assertEqual(zone_update['ttl'], 7200)
        self.assertEqual(zone_update['masters'], zone['masters'])

        # Test we can delete and get True returned
        zone_delete = self.user_cloud.delete_zone(zone['id'])
        self.assertTrue(zone_delete)

    def cleanup(self, name):
        self.user_cloud.delete_zone(name)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from keystoneauth1 import adapter

from openstack.baremetal.v1 import allocation
from openstack import exceptions
from openstack.tests.unit import base

FAKE = {
    "candidate_nodes": [],
    "created_at": "2016-08-18T22:28:48.165105+00:00",
    "extra": {},
    "last_error": None,
    "links": [
        {
            "href": "http://127.0.0.1:6385/v1/allocations/<PG_ID>",
            "rel": "self"
        },
        {
            "href": "http://127.0.0.1:6385/allocations/<PG_ID>",
            "rel": "bookmark"
        }
    ],
    "name": "test_allocation",
    "node_uuid": "6d85703a-565d-469a-96ce-30b6de53079d",
    "resource_class": "baremetal",
    "state": "active",
    "traits": [],
    "updated_at": None,
    "uuid": "e43c722c-248e-4c6e-8ce8-0d8ff129387a",
}


class TestAllocation(base.TestCase):

    def test_basic(self):
        sot = allocation.Allocation()
        self.assertIsNone(sot.resource_key)
        self.assertEqual('allocations', sot.resources_key)
        self.assertEqual('/allocations', sot.base_path)
        self.assertTrue(sot.allow_create)
        self.assertTrue(sot.allow_fetch)
        self.assertTrue(sot.allow_commit)
        self.assertTrue(sot.allow_delete)
        self.assertTrue(sot.allow_list)

    def test_instantiate(self):
        sot = allocation.Allocation(**FAKE)
        self.assertEqual(FAKE['candidate_nodes'], sot.candidate_nodes)
        self.assertEqual(FAKE['created_at'], sot.created_at)
        self.assertEqual(FAKE['extra'], sot.extra)
        self.assertEqual(FAKE['last_error'], sot.last_error)
        self.assertEqual(FAKE['links'], sot.links)
        self.assertEqual(FAKE['name'], sot.name)
        self.assertEqual(FAKE['node_uuid'], sot.node_id)
        self.assertEqual(FAKE['resource_class'], sot.resource_class)
        self.assertEqual(FAKE['state'], sot.state)
        self.assertEqual(FAKE['traits'], sot.traits)
        self.assertEqual(FAKE['updated_at'], sot.updated_at)
        self.assertEqual(FAKE['uuid'], sot.id)


@mock.patch('time.sleep', lambda _t: None)
@mock.patch.object(allocation.Allocation, 'fetch', autospec=True)
class TestWaitForAllocation(base.TestCase):

    def setUp(self):
        super(TestWaitForAllocation, self).setUp()
        self.session = mock.Mock(spec=adapter.Adapter)
        self.session.default_microversion = '1.52'
        self.session.log = mock.Mock()
        self.fake = dict(FAKE, state='allocating', node_uuid=None)
        self.allocation = allocation.Allocation(**self.fake)

    def test_already_active(self, mock_fetch):
        self.allocation.state = 'active'
        allocation = self.allocation.wait(None)
        self.assertIs(allocation, self.allocation)
        self.assertFalse(mock_fetch.called)

    def test_wait(self, mock_fetch):
        marker = [False]  # mutable object to modify in the closure

        def _side_effect(allocation, session):
            if marker[0]:
                self.allocation.state = 'active'
                self.allocation.node_id = FAKE['node_uuid']
            else:
                marker[0] = True

        mock_fetch.side_effect = _side_effect
        allocation = self.allocation.wait(self.session)
        self.assertIs(allocation, self.allocation)
        self.assertEqual(2, mock_fetch.call_count)

    def test_failure(self, mock_fetch):
        marker = [False]  # mutable object to modify in the closure

        def _side_effect(allocation, session):
            if marker[0]:
                self.allocation.state = 'error'
                self.allocation.last_error = 'boom!'
            else:
                marker[0] = True

        mock_fetch.side_effect = _side_effect
        self.assertRaises(exceptions.ResourceFailure,
                          self.allocation.wait, self.session)
        self.assertEqual(2, mock_fetch.call_count)

    def test_failure_ignored(self, mock_fetch):
        marker = [False]  # mutable object to modify in the closure

        def _side_effect(allocation, session):
            if marker[0]:
                self.allocation.state = 'error'
                self.allocation.last_error = 'boom!'
            else:
                marker[0] = True

        mock_fetch.side_effect = _side_effect
        allocation = self.allocation.wait(self.session, ignore_error=True)
        self.assertIs(allocation, self.allocation)
        self.assertEqual(2, mock_fetch.call_count)

    def test_timeout(self, mock_fetch):
        self.assertRaises(exceptions.ResourceTimeout,
                          self.allocation.wait, self.session, timeout=0.001)
        mock_fetch.assert_called_with(self.allocation, self.session)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.dns.v2 import _base
from openstack import resource


class ZoneTransferBase(_base.Resource):
    """DNS Zone Transfer Request/Accept Base Resource"""

    _query_mapping = resource.QueryParameters(
        'status'
    )

    #: Properties
    #: Timestamp when the resource was created
    created_at = resource.Body('created_at')
    #: Key that is used as part of the zone transfer accept process.
    #: This is only shown to the creator, and must be communicated out of band.
    key = resource.Body('key')
    #: The project id which the zone belongs to
    project_id = resource.Body('project_id')
    #: Current status of the zone import
    status = resource.Body('status')
    #: Timestamp when the resource was last updated
    updated_at = resource.Body('updated_at')
    #: Version of the resource
    version = resource.Body('version', type=int)
    #: ID for the zone that is being exported
    zone_id = resource.Body('zone_id')


class ZoneTransferRequest(ZoneTransferBase):
    """DNS Zone Transfer Request Resource"""
    base_path = '/zones/tasks/transfer_requests'
    resources_key = 'transfer_requests'

    # capabilities
    allow_create = True
    allow_fetch = True
    allow_delete = True
    allow_list = True
    allow_commit = True

    #: Description
    description = resource.Body('description')
    #: A project ID that the request will be limited to.
    #: No other project will be allowed to accept this request.
    target_project_id = resource.Body('target_project_id')
    #: Name for the zone that is being exported
    zone_name = resource.Body('zone_name')


class ZoneTransferAccept(ZoneTransferBase):
    """DNS Zone Transfer Accept Resource"""
    base_path = '/zones/tasks/transfer_accepts'
    resources_key = 'transfer_accepts'

    # capabilities
    allow_create = True
    allow_fetch = True
    allow_list = True

    #: Name for the zone that is being exported
    zone_transfer_request_id = resource.Body('zone_transfer_request_id')



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.shared_file_system.v2 import share
from openstack.tests.unit import base

IDENTIFIER = '08a87d37-5ca2-4308-86c5-cba06d8d796c'
EXAMPLE = {
    "id": IDENTIFIER,
    "size": 1,
    "availability_zone": "manila-zone-1",
    "created_at": "2021-02-11T17:38:00.000000",
    "status": "available",
    "name": None,
    "description": None,
    "project_id": "d19444eb73af4b37bc0794532ef6fc50",
    "snapshot_id": None,
    "share_network_id": None,
    "share_protocol": "NFS",
    "metadata": {},
    "share_type": "cbb18bb7-cc97-477a-b64b-ed7c7f2a1c67",
    "volume_type": "default",
    "is_public": False,
    "is_snapshot_supported": True,
    "task_state": None,
    "share_type_name": "default",
    "access_rules_status": "active",
    "replication_type": None,
    "is_replicated": False,
    "user_id": "6c262cab98de42c2afc4cfccbefc50c7",
    "is_creating_new_share_from_snapshot_supported": True,
    "is_reverting_to_snapshot_supported": True,
    "share_group_id": None,
    "source_share_group_snapshot_member_id": None,
    "is_mounting_snapshot_supported": True,
    "progress": "100%",
    "share_server_id": None,
    "host": "new@denver#lvm-single-pool"
}


class TestShares(base.TestCase):

    def test_basic(self):
        shares_resource = share.Share()
        self.assertEqual('shares', shares_resource.resources_key)
        self.assertEqual('/shares', shares_resource.base_path)
        self.assertTrue(shares_resource.allow_list)
        self.assertTrue(shares_resource.allow_create)
        self.assertTrue(shares_resource.allow_fetch)
        self.assertTrue(shares_resource.allow_commit)
        self.assertTrue(shares_resource.allow_delete)

    def test_make_shares(self):
        shares_resource = share.Share(**EXAMPLE)
        self.assertEqual(EXAMPLE['id'], shares_resource.id)
        self.assertEqual(EXAMPLE['size'], shares_resource.size)
        self.assertEqual(EXAMPLE['availability_zone'],
                         shares_resource.availability_zone)
        self.assertEqual(EXAMPLE['created_at'], shares_resource.created_at)
        self.assertEqual(EXAMPLE['status'], shares_resource.status)
        self.assertEqual(EXAMPLE['name'], shares_resource.name)
        self.assertEqual(EXAMPLE['description'],
                         shares_resource.description)
        self.assertEqual(EXAMPLE['project_id'], shares_resource.project_id)
        self.assertEqual(EXAMPLE['snapshot_id'], shares_resource.snapshot_id)
        self.assertEqual(EXAMPLE['share_network_id'],
                         shares_resource.share_network_id)
        self.assertEqual(EXAMPLE['share_protocol'],
                         shares_resource.share_protocol)
        self.assertEqual(EXAMPLE['metadata'], shares_resource.metadata)
        self.assertEqual(EXAMPLE['share_type'], shares_resource.share_type)
        self.assertEqual(EXAMPLE['is_public'], shares_resource.is_public)
        self.assertEqual(EXAMPLE['is_snapshot_supported'],
                         shares_resource.is_snapshot_supported)
        self.assertEqual(EXAMPLE['task_state'], shares_resource.task_state)
        self.assertEqual(EXAMPLE['share_type_name'],
                         shares_resource.share_type_name)
        self.assertEqual(EXAMPLE['access_rules_status'],
                         shares_resource.access_rules_status)
        self.assertEqual(EXAMPLE['replication_type'],
                         shares_resource.replication_type)
        self.assertEqual(EXAMPLE['is_replicated'],
                         shares_resource.is_replicated)
        self.assertEqual(EXAMPLE['user_id'], shares_resource.user_id)
        self.assertEqual(EXAMPLE[
            'is_creating_new_share_from_snapshot_supported'],
            (shares_resource.is_creating_new_share_from_snapshot_supported))
        self.assertEqual(EXAMPLE['is_reverting_to_snapshot_supported'],
                         shares_resource.is_reverting_to_snapshot_supported)
        self.assertEqual(EXAMPLE['share_group_id'],
                         shares_resource.share_group_id)
        self.assertEqual(EXAMPLE[
            'source_share_group_snapshot_member_id'],
            shares_resource.source_share_group_snapshot_member_id)
        self.assertEqual(EXAMPLE['is_mounting_snapshot_supported'],
                         shares_resource.is_mounting_snapshot_supported)
        self.assertEqual(EXAMPLE['progress'],
                         shares_resource.progress)
        self.assertEqual(EXAMPLE['share_server_id'],
                         shares_resource.share_server_id)
        self.assertEqual(EXAMPLE['host'], shares_resource.host)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from openstack import resource


class Action(resource.Resource):
    resource_key = 'action'
    resources_key = 'actions'
    base_path = '/actions'

    # Capabilities
    allow_list = True
    allow_fetch = True
    allow_commit = True

    commit_method = 'PATCH'

    _query_mapping = resource.QueryParameters(
        'name', 'action', 'status', 'sort', 'global_project',
        'cluster_id', target_id='target')

    # Properties
    #: Name of the action.
    name = resource.Body('name')
    #: ID of the target object, which can be a cluster or a node.
    target_id = resource.Body('target')
    #: Built-in type name of action.
    action = resource.Body('action')
    #: A string representation of the reason why the action was created.
    cause = resource.Body('cause')
    #: The owning engine that is currently running the action.
    owner_id = resource.Body('owner')
    #: The ID of the user who created this action.
    user_id = resource.Body('user')
    #: The ID of the project this profile belongs to.
    project_id = resource.Body('project')
    #: The domain ID of the action.
    domain_id = resource.Body('domain')
    #: Interval in seconds between two consecutive executions.
    interval = resource.Body('interval')
    #: The time the action was started.
    start_at = resource.Body('start_time')
    #: The time the action completed execution.
    end_at = resource.Body('end_time')
    #: The timeout in seconds.
    timeout = resource.Body('timeout')
    #: Current status of the action.
    status = resource.Body('status')
    #: A string describing the reason that brought the action to its current
    #  status.
    status_reason = resource.Body('status_reason')
    #: A dictionary containing the inputs to the action.
    inputs = resource.Body('inputs', type=dict)
    #: A dictionary containing the outputs to the action.
    outputs = resource.Body('outputs', type=dict)
    #: A list of actions that must finish before this action starts execution.
    depends_on = resource.Body('depends_on', type=list)
    #: A list of actions that can start only after this action has finished.
    depended_by = resource.Body('depended_by', type=list)
    #: Timestamp when the action is created.
    created_at = resource.Body('created_at')
    #: Timestamp when the action was last updated.
    updated_at = resource.Body('updated_at')
    #: The ID of cluster which this action runs on.
    cluster_id = resource.Body('cluster_id')



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
test_quotas
----------------------------------

Functional tests for `shade` quotas methods.
"""

from openstack.tests.functional import base


class TestComputeQuotas(base.BaseFunctionalTest):

    def test_quotas(self):
        '''Test quotas functionality'''
        quotas = self.operator_cloud.get_compute_quotas('demo')
        cores = quotas['cores']
        self.operator_cloud.set_compute_quotas('demo', cores=cores + 1)
        self.assertEqual(
            cores + 1,
            self.operator_cloud.get_compute_quotas('demo')['cores'])
        self.operator_cloud.delete_compute_quotas('demo')
        self.assertEqual(
            cores, self.operator_cloud.get_compute_quotas('demo')['cores'])


class TestVolumeQuotas(base.BaseFunctionalTest):

    def setUp(self):
        super(TestVolumeQuotas, self).setUp()
        if not self.operator_cloud.has_service('volume'):
            self.skipTest('volume service not supported by cloud')

    def test_quotas(self):
        '''Test quotas functionality'''
        quotas = self.operator_cloud.get_volume_quotas('demo')
        volumes = quotas['volumes']
        self.operator_cloud.set_volume_quotas('demo', volumes=volumes + 1)
        self.assertEqual(
            volumes + 1,
            self.operator_cloud.get_volume_quotas('demo')['volumes'])
        self.operator_cloud.delete_volume_quotas('demo')
        self.assertEqual(
            volumes,
            self.operator_cloud.get_volume_quotas('demo')['volumes'])


class TestNetworkQuotas(base.BaseFunctionalTest):

    def setUp(self):
        super(TestNetworkQuotas, self).setUp()
        if not self.operator_cloud.has_service('network'):
            self.skipTest('network service not supported by cloud')

    def test_quotas(self):
        '''Test quotas functionality'''
        quotas = self.operator_cloud.get_network_quotas('demo')
        network = quotas['networks']
        self.operator_cloud.set_network_quotas('demo', networks=network + 1)
        self.assertEqual(
            network + 1,
            self.operator_cloud.get_network_quotas('demo')['networks'])
        self.operator_cloud.delete_network_quotas('demo')
        self.assertEqual(
            network,
            self.operator_cloud.get_network_quotas('demo')['networks'])

    def test_get_quotas_details(self):
        quotas = [
            'floating_ips', 'networks', 'ports',
            'rbac_policies', 'routers', 'subnets',
            'subnet_pools', 'security_group_rules',
            'security_groups']
        expected_keys = ['limit', 'used', 'reserved']
        '''Test getting details about quota usage'''
        quota_details = self.operator_cloud.get_network_quotas(
            'demo', details=True)
        for quota in quotas:
            quota_val = quota_details[quota]
            if quota_val:
                for expected_key in expected_keys:
                    self.assertTrue(expected_key in quota_val)



# Copyright 2018 Red Hat, Inc.
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import warnings

import os_service_types

from openstack import _log
from openstack import exceptions
from openstack import proxy as proxy_mod

__all__ = [
    'ServiceDescription',
]

_logger = _log.setup_logging('openstack')
_service_type_manager = os_service_types.ServiceTypes()


class _ServiceDisabledProxyShim:
    def __init__(self, service_type, reason):
        self.service_type = service_type
        self.reason = reason

    def __getattr__(self, item):
        raise exceptions.ServiceDisabledException(
            "Service '{service_type}' is disabled because its configuration "
            "could not be loaded. {reason}".format(
                service_type=self.service_type, reason=self.reason or ''))


class ServiceDescription:

    #: Dictionary of supported versions and proxy classes for that version
    supported_versions = None
    #: main service_type to use to find this service in the catalog
    service_type = None
    #: list of aliases this service might be registered as
    aliases = []

    def __init__(self, service_type, supported_versions=None, aliases=None):
        """Class describing how to interact with a REST service.

        Each service in an OpenStack cloud needs to be found by looking
        for it in the catalog. Once the endpoint is found, REST calls can
        be made, but a Proxy class and some Resource objects are needed
        to provide an object interface.

        Instances of ServiceDescription can be passed to
        `openstack.connection.Connection.add_service`, or a list can be
        passed to the `openstack.connection.Connection` constructor in
        the ``extra_services`` argument.

        All three parameters can be provided at instantation time, or
        a service-specific subclass can be used that sets the attributes
        directly.

        :param string service_type:
            service_type to look for in the keystone catalog
        :param list aliases:
            Optional list of aliases, if there is more than one name that might
            be used to register the service in the catalog.
        """
        self.service_type = service_type or self.service_type
        self.supported_versions = (
            supported_versions
            or self.supported_versions
            or {})

        self.aliases = aliases or self.aliases
        self.all_types = [service_type] + self.aliases

    def __get__(self, instance, owner):
        if instance is None:
            return self
        if self.service_type not in instance._proxies:
            proxy = self._make_proxy(instance)
            if not isinstance(proxy, _ServiceDisabledProxyShim):
                # The keystone proxy has a method called get_endpoint
                # that is about managing keystone endpoints. This is
                # unfortunate.
                try:
                    endpoint = proxy_mod.Proxy.get_endpoint(proxy)
                except IndexError:
                    # It's best not to look to closely here. This is
                    # to support old placement.
                    # There was a time when it had no status entry
                    # in its version discovery doc (OY) In this case,
                    # no endpoints get through version discovery
                    # filtering. In order to deal with that, catch
                    # the IndexError thrown by keystoneauth and
                    # set an endpoint_override for the user to the
                    # url in the catalog and try again.
                    self._set_override_from_catalog(instance.config)
                    proxy = self._make_proxy(instance)
                    endpoint = proxy_mod.Proxy.get_endpoint(proxy)
                if instance._strict_proxies:
                    self._validate_proxy(proxy, endpoint)
                proxy._connection = instance
            instance._proxies[self.service_type] = proxy
        return instance._proxies[self.service_type]

    def _set_override_from_catalog(self, config):
        override = config._get_endpoint_from_catalog(
            self.service_type,
            proxy_mod.Proxy,
        )
        config.set_service_value(
            'endpoint_override',
            self.service_type,
            override,
        )

    def _validate_proxy(self, proxy, endpoint):
        exc = None
        service_url = getattr(proxy, 'skip_discovery', None)
        try:
            # Don't go too wild for e.g. swift
            if service_url is None:
                service_url = proxy.get_endpoint_data().service_url
        except Exception as e:
            exc = e
        if exc or not endpoint or not service_url:
            raise exceptions.ServiceDiscoveryException(
                "Failed to create a working proxy for service {service_type}: "
                "{message}".format(
                    service_type=self.service_type,
                    message=exc or "No valid endpoint was discoverable."))

    def _make_proxy(self, instance):
        """Create a Proxy for the service in question.

        :param instance:
          The `openstack.connection.Connection` we're working with.
        """
        config = instance.config

        if not config.has_service(self.service_type):
            return _ServiceDisabledProxyShim(
                self.service_type,
                config.get_disabled_reason(self.service_type))

        # We don't know anything about this service, so the user is
        # explicitly just using us for a passthrough REST adapter.
        # Skip all the lower logic.
        if not self.supported_versions:
            temp_client = config.get_session_client(
                self.service_type,
                allow_version_hack=True,
            )
            return temp_client

        # Check to see if we've got config that matches what we
        # understand in the SDK.
        version_string = config.get_api_version(self.service_type)
        endpoint_override = config.get_endpoint(self.service_type)

        # If the user doesn't give a version in config, but we only support
        # one version, then just use that version.
        if not version_string and len(self.supported_versions) == 1:
            version_string = list(self.supported_versions)[0]

        proxy_obj = None
        if endpoint_override and version_string:
            # Both endpoint override and version_string are set, we don't
            # need to do discovery - just trust the user.
            proxy_class = self.supported_versions.get(version_string[0])
            if proxy_class:
                proxy_obj = config.get_session_client(
                    self.service_type,
                    constructor=proxy_class,
                )
            else:
                warnings.warn(
                    "The configured version, {version} for service"
                    " {service_type} is not known or supported by"
                    " openstacksdk. The resulting Proxy object will only"
                    " have direct passthrough REST capabilities.".format(
                        version=version_string,
                        service_type=self.service_type),
                    category=exceptions.UnsupportedServiceVersion)
        elif endpoint_override:
            temp_adapter = config.get_session_client(
                self.service_type
            )
            api_version = temp_adapter.get_endpoint_data().api_version
            proxy_class = self.supported_versions.get(str(api_version[0]))
            if proxy_class:
                proxy_obj = config.get_session_client(
                    self.service_type,
                    constructor=proxy_class,
                )
            else:
                warnings.warn(
                    "Service {service_type} has an endpoint override set"
                    " but the version discovered at that endpoint, {version}"
                    " is not supported by openstacksdk. The resulting Proxy"
                    " object will only have direct passthrough REST"
                    " capabilities.".format(
                        version=api_version,
                        service_type=self.service_type),
                    category=exceptions.UnsupportedServiceVersion)

        if proxy_obj:

            if getattr(proxy_obj, 'skip_discovery', False):
                # Some services, like swift, don't have discovery. While
                # keystoneauth will behave correctly and handle such
                # scenarios, it's not super efficient as it involves trying
                # and falling back a few times.
                return proxy_obj

            data = proxy_obj.get_endpoint_data()
            if not data and instance._strict_proxies:
                raise exceptions.ServiceDiscoveryException(
                    "Failed to create a working proxy for service "
                    "{service_type}: No endpoint data found.".format(
                        service_type=self.service_type))

            # If we've gotten here with a proxy object it means we have
            # an endpoint_override in place. If the catalog_url and
            # service_url don't match, which can happen if there is a
            # None plugin and auth.endpoint like with standalone ironic,
            # we need to be explicit that this service has an endpoint_override
            # so that subsequent discovery calls don't get made incorrectly.
            if data.catalog_url != data.service_url:
                ep_key = '{service_type}_endpoint_override'.format(
                    service_type=self.service_type.replace('-', '_'))
                config.config[ep_key] = data.service_url
                proxy_obj = config.get_session_client(
                    self.service_type,
                    constructor=proxy_class,
                )
            return proxy_obj

        # Make an adapter to let discovery take over
        version_kwargs = {}
        if version_string:
            version_kwargs['version'] = version_string
        else:
            supported_versions = sorted([
                int(f) for f in self.supported_versions])
            version_kwargs['min_version'] = str(supported_versions[0])
            version_kwargs['max_version'] = '{version}.latest'.format(
                version=str(supported_versions[-1]))

        temp_adapter = config.get_session_client(
            self.service_type,
            allow_version_hack=True,
            **version_kwargs
        )
        found_version = temp_adapter.get_api_major_version()
        if found_version is None:
            region_name = instance.config.get_region_name(self.service_type)
            if version_kwargs:
                raise exceptions.NotSupported(
                    "The {service_type} service for {cloud}:{region_name}"
                    " exists but does not have any supported versions.".format(
                        service_type=self.service_type,
                        cloud=instance.name,
                        region_name=region_name))
            else:
                raise exceptions.NotSupported(
                    "The {service_type} service for {cloud}:{region_name}"
                    " exists but no version was discoverable.".format(
                        service_type=self.service_type,
                        cloud=instance.name,
                        region_name=region_name))
        proxy_class = self.supported_versions.get(str(found_version[0]))
        if proxy_class:
            return config.get_session_client(
                self.service_type,
                allow_version_hack=True,
                constructor=proxy_class,
                **version_kwargs
            )

        # No proxy_class
        # Maybe openstacksdk is being used for the passthrough
        # REST API proxy layer for an unknown service in the
        # service catalog that also doesn't have any useful
        # version discovery?
        warnings.warn(
            "Service {service_type} has no discoverable version."
            " The resulting Proxy object will only have direct"
            " passthrough REST capabilities.".format(
                service_type=self.service_type),
            category=exceptions.UnsupportedServiceVersion)
        return temp_adapter

    def __set__(self, instance, value):
        raise AttributeError('Service Descriptors cannot be set')

    def __delete__(self, instance):
        # NOTE(gtema) Some clouds are not very fast (or interested at all)
        # in bringing their changes upstream. If there are incompatible changes
        # downstream we need to allow overriding default implementation by
        # deleting service_type attribute of the connection and then
        # "add_service" with new implementation.
        # This is implemented explicitely not very comfortable to use
        # to show how bad it is not to contribute changes back
        for service_type in self.all_types:
            if service_type in instance._proxies:
                del instance._proxies[service_type]



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# import types so that we can reference ListType in sphinx param declarations.
# We can't just use list, because sphinx gets confused by
# openstack.resource.Resource.list and openstack.resource2.Resource.list
import threading
import types  # noqa

from openstack.cloud import exc


class NetworkCommonCloudMixin:
    """Shared networking functions used by FloatingIP, Network, Compute classes
    """

    def __init__(self):
        self._external_ipv4_names = self.config.get_external_ipv4_networks()
        self._internal_ipv4_names = self.config.get_internal_ipv4_networks()
        self._external_ipv6_names = self.config.get_external_ipv6_networks()
        self._internal_ipv6_names = self.config.get_internal_ipv6_networks()
        self._nat_destination = self.config.get_nat_destination()
        self._nat_source = self.config.get_nat_source()
        self._default_network = self.config.get_default_network()

        self._use_external_network = self.config.config.get(
            'use_external_network', True)
        self._use_internal_network = self.config.config.get(
            'use_internal_network', True)

        self._networks_lock = threading.Lock()
        self._reset_network_caches()

    def use_external_network(self):
        return self._use_external_network

    def use_internal_network(self):
        return self._use_internal_network

    def _reset_network_caches(self):
        # Variables to prevent us from going through the network finding
        # logic again if we've done it once. This is different from just
        # the cached value, since "None" is a valid value to find.
        with self._networks_lock:
            self._external_ipv4_networks = []
            self._external_ipv4_floating_networks = []
            self._internal_ipv4_networks = []
            self._external_ipv6_networks = []
            self._internal_ipv6_networks = []
            self._nat_destination_network = None
            self._nat_source_network = None
            self._default_network_network = None
            self._network_list_stamp = False

    def _set_interesting_networks(self):
        external_ipv4_networks = []
        external_ipv4_floating_networks = []
        internal_ipv4_networks = []
        external_ipv6_networks = []
        internal_ipv6_networks = []
        nat_destination = None
        nat_source = None
        default_network = None

        all_subnets = None

        # Filter locally because we have an or condition
        try:
            # TODO(mordred): Rackspace exposes neutron but it does not
            # work. I think that overriding what the service catalog
            # reports should be a thing os-client-config should handle
            # in a vendor profile - but for now it does not. That means
            # this search_networks can just totally fail. If it does
            # though, that's fine, clearly the neutron introspection is
            # not going to work.
            all_networks = self.list_networks()
        except exc.OpenStackCloudException:
            self._network_list_stamp = True
            return

        for network in all_networks:

            # External IPv4 networks
            if (network['name'] in self._external_ipv4_names
                    or network['id'] in self._external_ipv4_names):
                external_ipv4_networks.append(network)
            elif ((network.is_router_external
                   or network.provider_physical_network)
                  and network['name'] not in self._internal_ipv4_names
                  and network['id'] not in self._internal_ipv4_names):
                external_ipv4_networks.append(network)

            # Internal networks
            if (network['name'] in self._internal_ipv4_names
                    or network['id'] in self._internal_ipv4_names):
                internal_ipv4_networks.append(network)
            elif (not network.is_router_external
                  and not network.provider_physical_network
                  and network['name'] not in self._external_ipv4_names
                  and network['id'] not in self._external_ipv4_names):
                internal_ipv4_networks.append(network)

            # External networks
            if (network['name'] in self._external_ipv6_names
                    or network['id'] in self._external_ipv6_names):
                external_ipv6_networks.append(network)
            elif (network.is_router_external
                  and network['name'] not in self._internal_ipv6_names
                  and network['id'] not in self._internal_ipv6_names):
                external_ipv6_networks.append(network)

            # Internal networks
            if (network['name'] in self._internal_ipv6_names
                    or network['id'] in self._internal_ipv6_names):
                internal_ipv6_networks.append(network)
            elif (not network.is_router_external
                  and network['name'] not in self._external_ipv6_names
                  and network['id'] not in self._external_ipv6_names):
                internal_ipv6_networks.append(network)

            # External Floating IPv4 networks
            if self._nat_source in (
                    network['name'], network['id']):
                if nat_source:
                    raise exc.OpenStackCloudException(
                        'Multiple networks were found matching'
                        ' {nat_net} which is the network configured'
                        ' to be the NAT source. Please check your'
                        ' cloud resources. It is probably a good idea'
                        ' to configure this network by ID rather than'
                        ' by name.'.format(
                            nat_net=self._nat_source))
                external_ipv4_floating_networks.append(network)
                nat_source = network
            elif self._nat_source is None:
                if network.is_router_external:
                    external_ipv4_floating_networks.append(network)
                    nat_source = nat_source or network

            # NAT Destination
            if self._nat_destination in (
                    network['name'], network['id']):
                if nat_destination:
                    raise exc.OpenStackCloudException(
                        'Multiple networks were found matching'
                        ' {nat_net} which is the network configured'
                        ' to be the NAT destination. Please check your'
                        ' cloud resources. It is probably a good idea'
                        ' to configure this network by ID rather than'
                        ' by name.'.format(
                            nat_net=self._nat_destination))
                nat_destination = network
            elif self._nat_destination is None:
                # TODO(mordred) need a config value for floating
                # ips for this cloud so that we can skip this
                # No configured nat destination, we have to figured
                # it out.
                if all_subnets is None:
                    try:
                        all_subnets = self.list_subnets()
                    except exc.OpenStackCloudException:
                        # Thanks Rackspace broken neutron
                        all_subnets = []

                for subnet in all_subnets:
                    # TODO(mordred) trap for detecting more than
                    # one network with a gateway_ip without a config
                    if ('gateway_ip' in subnet and subnet['gateway_ip']
                            and network['id'] == subnet['network_id']):
                        nat_destination = network
                        break

            # Default network
            if self._default_network in (
                    network['name'], network['id']):
                if default_network:
                    raise exc.OpenStackCloudException(
                        'Multiple networks were found matching'
                        ' {default_net} which is the network'
                        ' configured to be the default interface'
                        ' network. Please check your cloud resources.'
                        ' It is probably a good idea'
                        ' to configure this network by ID rather than'
                        ' by name.'.format(
                            default_net=self._default_network))
                default_network = network

        # Validate config vs. reality
        for net_name in self._external_ipv4_names:
            if net_name not in [net['name'] for net in external_ipv4_networks]:
                raise exc.OpenStackCloudException(
                    "Networks: {network} was provided for external IPv4"
                    " access and those networks could not be found".format(
                        network=net_name))

        for net_name in self._internal_ipv4_names:
            if net_name not in [net['name'] for net in internal_ipv4_networks]:
                raise exc.OpenStackCloudException(
                    "Networks: {network} was provided for internal IPv4"
                    " access and those networks could not be found".format(
                        network=net_name))

        for net_name in self._external_ipv6_names:
            if net_name not in [net['name'] for net in external_ipv6_networks]:
                raise exc.OpenStackCloudException(
                    "Networks: {network} was provided for external IPv6"
                    " access and those networks could not be found".format(
                        network=net_name))

        for net_name in self._internal_ipv6_names:
            if net_name not in [net['name'] for net in internal_ipv6_networks]:
                raise exc.OpenStackCloudException(
                    "Networks: {network} was provided for internal IPv6"
                    " access and those networks could not be found".format(
                        network=net_name))

        if self._nat_destination and not nat_destination:
            raise exc.OpenStackCloudException(
                'Network {network} was configured to be the'
                ' destination for inbound NAT but it could not be'
                ' found'.format(
                    network=self._nat_destination))

        if self._nat_source and not nat_source:
            raise exc.OpenStackCloudException(
                'Network {network} was configured to be the'
                ' source for inbound NAT but it could not be'
                ' found'.format(
                    network=self._nat_source))

        if self._default_network and not default_network:
            raise exc.OpenStackCloudException(
                'Network {network} was configured to be the'
                ' default network interface but it could not be'
                ' found'.format(
                    network=self._default_network))

        self._external_ipv4_networks = external_ipv4_networks
        self._external_ipv4_floating_networks = external_ipv4_floating_networks
        self._internal_ipv4_networks = internal_ipv4_networks
        self._external_ipv6_networks = external_ipv6_networks
        self._internal_ipv6_networks = internal_ipv6_networks
        self._nat_destination_network = nat_destination
        self._nat_source_network = nat_source
        self._default_network_network = default_network

    def _find_interesting_networks(self):
        if self._networks_lock.acquire():
            try:
                if self._network_list_stamp:
                    return
                if (
                    not self._use_external_network
                    and not self._use_internal_network
                ):
                    # Both have been flagged as skip - don't do a list
                    return
                if not self.has_service('network'):
                    return
                self._set_interesting_networks()
                self._network_list_stamp = True
            finally:
                self._networks_lock.release()

    def get_nat_destination(self):
        """Return the network that is configured to be the NAT destination.

        :returns: A network ``Network`` object if one is found
        """
        self._find_interesting_networks()
        return self._nat_destination_network

    def get_nat_source(self):
        """Return the network that is configured to be the NAT destination.

        :returns: A network ``Network`` object if one is found
        """
        self._find_interesting_networks()
        return self._nat_source_network

    def get_default_network(self):
        """Return the network that is configured to be the default interface.

        :returns: A network ``Network`` object if one is found
        """
        self._find_interesting_networks()
        return self._default_network_network

    def get_external_networks(self):
        """Return the networks that are configured to route northbound.

        This should be avoided in favor of the specific ipv4/ipv6 method,
        but is here for backwards compatibility.

        :returns: A list of network ``Network`` objects if any are found
        """
        self._find_interesting_networks()
        return (
            list(self._external_ipv4_networks)
            + list(self._external_ipv6_networks)
        )

    def get_internal_networks(self):
        """Return the networks that are configured to not route northbound.

        This should be avoided in favor of the specific ipv4/ipv6 method,
        but is here for backwards compatibility.

        :returns: A list of network ``Network`` objects if any are found
        """
        self._find_interesting_networks()
        return (
            list(self._internal_ipv4_networks)
            + list(self._internal_ipv6_networks)
        )

    def get_external_ipv4_networks(self):
        """Return the networks that are configured to route northbound.

        :returns: A list of network ``Network`` objects if any are found
        """
        self._find_interesting_networks()
        return self._external_ipv4_networks

    def get_external_ipv4_floating_networks(self):
        """Return the networks that are configured to route northbound.

        :returns: A list of network ``Network`` objects if any are found
        """
        self._find_interesting_networks()
        return self._external_ipv4_floating_networks

    def get_internal_ipv4_networks(self):
        """Return the networks that are configured to not route northbound.

        :returns: A list of network ``Network`` objects if any are found
        """
        self._find_interesting_networks()
        return self._internal_ipv4_networks

    def get_external_ipv6_networks(self):
        """Return the networks that are configured to route northbound.

        :returns: A list of network ``Network`` objects if any are found
        """
        self._find_interesting_networks()
        return self._external_ipv6_networks

    def get_internal_ipv6_networks(self):
        """Return the networks that are configured to not route northbound.

        :returns: A list of network ``Network`` objects if any are found
        """
        self._find_interesting_networks()
        return self._internal_ipv6_networks



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.placement.v1 import _proxy
from openstack.placement.v1 import resource_class
from openstack.placement.v1 import resource_provider
from openstack.tests.unit import test_proxy_base as test_proxy_base


class TestPlacementProxy(test_proxy_base.TestProxyBase):

    def setUp(self):
        super().setUp()
        self.proxy = _proxy.Proxy(self.session)


# resource classes
class TestPlacementResourceClass:
    def test_resource_class_create(self):
        self.verify_create(
            self.proxy.create_resource_class,
            resource_class.ResourceClass,
        )

    def test_resource_class_delete(self):
        self.verify_delete(
            self.proxy.delete_resource_class,
            resource_class.ResourceClass,
            False,
        )

    def test_resource_class_update(self):
        self.verify_update(
            self.proxy.update_resource_class,
            resource_class.ResourceClass,
            False,
        )

    def test_resource_class_get(self):
        self.verify_get(
            self.proxy.get_resource_class,
            resource_class.ResourceClass,
        )

    def test_resource_classes(self):
        self.verify_list(
            self.proxy.resource_classes,
            resource_class.ResourceClass,
        )


# resource providers
class TestPlacementResourceProvider:
    def test_resource_provider_create(self):
        self.verify_create(
            self.proxy.create_resource_provider,
            resource_provider.ResourceProvider,
        )

    def test_resource_provider_delete(self):
        self.verify_delete(
            self.proxy.delete_resource_provider,
            resource_provider.ResourceProvider,
            False,
        )

    def test_resource_provider_update(self):
        self.verify_update(
            self.proxy.update_resource_provider,
            resource_provider.ResourceProvider,
            False,
        )

    def test_resource_provider_get(self):
        self.verify_get(
            self.proxy.get_resource_provider,
            resource_provider.ResourceProvider,
        )

    def test_resource_providers(self):
        self.verify_list(
            self.proxy.resource_providers,
            resource_provider.ResourceProvider,
        )



# Copyright 2019 Rackspace, US Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import uuid

from openstack.load_balancer.v2 import flavor_profile
from openstack.tests.unit import base


IDENTIFIER = uuid.uuid4()
EXAMPLE = {
    'id': IDENTIFIER,
    'name': 'acidic',
    'provider_name': 'best',
    'flavor_data': '{"loadbalancer_topology": "SINGLE"}'}


class TestFlavorProfile(base.TestCase):

    def test_basic(self):
        test_profile = flavor_profile.FlavorProfile()
        self.assertEqual('flavorprofile', test_profile.resource_key)
        self.assertEqual('flavorprofiles', test_profile.resources_key)
        self.assertEqual('/lbaas/flavorprofiles', test_profile.base_path)
        self.assertTrue(test_profile.allow_create)
        self.assertTrue(test_profile.allow_fetch)
        self.assertTrue(test_profile.allow_commit)
        self.assertTrue(test_profile.allow_delete)
        self.assertTrue(test_profile.allow_list)

    def test_make_it(self):
        test_profile = flavor_profile.FlavorProfile(**EXAMPLE)
        self.assertEqual(EXAMPLE['id'], test_profile.id)
        self.assertEqual(EXAMPLE['name'], test_profile.name)
        self.assertEqual(EXAMPLE['provider_name'], test_profile.provider_name)
        self.assertEqual(EXAMPLE['flavor_data'], test_profile.flavor_data)

        self.assertDictEqual(
            {'limit': 'limit',
             'marker': 'marker',
             'id': 'id',
             'name': 'name',
             'provider_name': 'provider_name',
             'flavor_data': 'flavor_data'},
            test_profile._query_mapping._mapping)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from keystoneauth1 import adapter

from openstack.baremetal.v1 import node as _node
from openstack.baremetal_introspection.v1 import _proxy
from openstack.baremetal_introspection.v1 import introspection
from openstack import exceptions
from openstack.tests.unit import base
from openstack.tests.unit import test_proxy_base


@mock.patch.object(introspection.Introspection, 'create', autospec=True)
class TestStartIntrospection(base.TestCase):

    def setUp(self):
        super(TestStartIntrospection, self).setUp()
        self.session = mock.Mock(spec=adapter.Adapter)
        self.proxy = _proxy.Proxy(self.session)

    def test_create_introspection(self, mock_create):
        self.proxy.start_introspection('abcd')
        mock_create.assert_called_once_with(mock.ANY, self.proxy)
        introspect = mock_create.call_args[0][0]
        self.assertEqual('abcd', introspect.id)

    def test_create_introspection_with_node(self, mock_create):
        self.proxy.start_introspection(_node.Node(id='abcd'))
        mock_create.assert_called_once_with(mock.ANY, self.proxy)
        introspect = mock_create.call_args[0][0]
        self.assertEqual('abcd', introspect.id)

    def test_create_introspection_manage_boot(self, mock_create):
        self.proxy.start_introspection('abcd', manage_boot=False)
        mock_create.assert_called_once_with(mock.ANY, self.proxy,
                                            manage_boot=False)
        introspect = mock_create.call_args[0][0]
        self.assertEqual('abcd', introspect.id)


class TestBaremetalIntrospectionProxy(test_proxy_base.TestProxyBase):

    def setUp(self):
        super(TestBaremetalIntrospectionProxy, self).setUp()
        self.proxy = _proxy.Proxy(self.session)

    def test_get_introspection(self):
        self.verify_get(self.proxy.get_introspection,
                        introspection.Introspection)


@mock.patch('time.sleep', lambda _sec: None)
@mock.patch.object(introspection.Introspection, 'fetch', autospec=True)
class TestWaitForIntrospection(base.TestCase):

    def setUp(self):
        super(TestWaitForIntrospection, self).setUp()
        self.session = mock.Mock(spec=adapter.Adapter)
        self.proxy = _proxy.Proxy(self.session)
        self.fake = {'state': 'waiting', 'error': None, 'finished': False}
        self.introspection = introspection.Introspection(**self.fake)

    def test_already_finished(self, mock_fetch):
        self.introspection.is_finished = True
        self.introspection.state = 'finished'
        result = self.proxy.wait_for_introspection(self.introspection)
        self.assertIs(result, self.introspection)
        self.assertFalse(mock_fetch.called)

    def test_wait(self, mock_fetch):
        marker = [False]  # mutable object to modify in the closure

        def _side_effect(allocation, session):
            if marker[0]:
                self.introspection.state = 'finished'
                self.introspection.is_finished = True
            else:
                self.introspection.state = 'processing'
                marker[0] = True

        mock_fetch.side_effect = _side_effect
        result = self.proxy.wait_for_introspection(self.introspection)
        self.assertIs(result, self.introspection)
        self.assertEqual(2, mock_fetch.call_count)

    def test_timeout(self, mock_fetch):
        self.assertRaises(exceptions.ResourceTimeout,
                          self.proxy.wait_for_introspection,
                          self.introspection,
                          timeout=0.001)
        mock_fetch.assert_called_with(self.introspection, self.proxy)

    def test_failure(self, mock_fetch):
        def _side_effect(allocation, session):
            self.introspection.state = 'error'
            self.introspection.is_finished = True
            self.introspection.error = 'boom'

        mock_fetch.side_effect = _side_effect
        self.assertRaisesRegex(exceptions.ResourceFailure, 'boom',
                               self.proxy.wait_for_introspection,
                               self.introspection)
        mock_fetch.assert_called_once_with(self.introspection, self.proxy)

    def test_failure_ignored(self, mock_fetch):
        def _side_effect(allocation, session):
            self.introspection.state = 'error'
            self.introspection.is_finished = True
            self.introspection.error = 'boom'

        mock_fetch.side_effect = _side_effect
        result = self.proxy.wait_for_introspection(self.introspection,
                                                   ignore_error=True)
        self.assertIs(result, self.introspection)
        mock_fetch.assert_called_once_with(self.introspection, self.proxy)


@mock.patch.object(_proxy.Proxy, 'request', autospec=True)
class TestAbortIntrospection(base.TestCase):

    def setUp(self):
        super(TestAbortIntrospection, self).setUp()
        self.session = mock.Mock(spec=adapter.Adapter)
        self.proxy = _proxy.Proxy(self.session)
        self.fake = {'id': '1234', 'finished': False}
        self.introspection = introspection.Introspection(**self.fake)

    def test_abort(self, mock_request):
        mock_request.return_value.status_code = 202
        self.proxy.abort_introspection(self.introspection)
        mock_request.assert_called_once_with(
            self.proxy, 'introspection/1234/abort', 'POST',
            headers=mock.ANY, microversion=mock.ANY,
            retriable_status_codes=[409, 503])


@mock.patch.object(_proxy.Proxy, 'request', autospec=True)
class TestGetData(base.TestCase):

    def setUp(self):
        super(TestGetData, self).setUp()
        self.session = mock.Mock(spec=adapter.Adapter)
        self.proxy = _proxy.Proxy(self.session)
        self.fake = {'id': '1234', 'finished': False}
        self.introspection = introspection.Introspection(**self.fake)

    def test_get_data(self, mock_request):
        mock_request.return_value.status_code = 200
        data = self.proxy.get_introspection_data(self.introspection)
        mock_request.assert_called_once_with(
            self.proxy, 'introspection/1234/data', 'GET',
            headers=mock.ANY, microversion=mock.ANY)
        self.assertIs(data, mock_request.return_value.json.return_value)

    def test_get_unprocessed_data(self, mock_request):
        mock_request.return_value.status_code = 200
        data = self.proxy.get_introspection_data(self.introspection,
                                                 processed=False)
        mock_request.assert_called_once_with(
            self.proxy, 'introspection/1234/data/unprocessed', 'GET',
            headers=mock.ANY, microversion='1.17')
        self.assertIs(data, mock_request.return_value.json.return_value)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack import resource


class ShareSnapshot(resource.Resource):
    resource_key = "snapshot"
    resources_key = "snapshots"
    base_path = "/snapshots"

    # capabilities
    allow_create = True
    allow_fetch = True
    allow_commit = True
    allow_delete = True
    allow_list = True
    allow_head = False

    _query_mapping = resource.QueryParameters(
        "snapshot_id"
    )

    #: Properties
    #: The date and time stamp when the resource was
    #: created within the service’s database.
    created_at = resource.Body("created_at")
    #: The user defined description of the resource.
    description = resource.Body("description", type=str)
    #: The user defined name of the resource.
    display_name = resource.Body("display_name", type=str)
    #: The user defined description of the resource
    display_description = resource.Body(
        "display_description", type=str)
    #: ID of the project that the snapshot belongs to.
    project_id = resource.Body("project_id", type=str)
    #: The UUID of the source share that was used to
    #: create the snapshot.
    share_id = resource.Body("share_id", type=str)
    #: The file system protocol of a share snapshot
    share_proto = resource.Body("share_proto", type=str)
    #: The snapshot's source share's size, in GiBs.
    share_size = resource.Body("share_size", type=int)
    #: The snapshot size, in GiBs.
    size = resource.Body("size", type=int)
    #: The snapshot status
    status = resource.Body("status", type=str)
    #: ID of the user that the snapshot was created by.
    user_id = resource.Body("user_id", type=str)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack import exceptions
from openstack import resource
from openstack import utils


class Group(resource.Resource):
    resource_key = "group"
    resources_key = "groups"
    base_path = "/groups"

    # capabilities
    allow_fetch = True
    allow_create = True
    allow_delete = True
    allow_commit = True
    allow_list = True

    availability_zone = resource.Body("availability_zone")
    created_at = resource.Body("created_at")
    description = resource.Body("description")
    group_snapshot_id = resource.Body("group_snapshot_id")
    group_type = resource.Body("group_type")
    project_id = resource.Body("project_id")
    replication_status = resource.Body("replication_status")
    source_group_id = resource.Body("source_group_id")
    status = resource.Body("status")
    volumes = resource.Body("volumes", type=list)
    volume_types = resource.Body("volume_types", type=list)

    _max_microversion = "3.38"

    def _action(self, session, body):
        """Preform group actions given the message body."""
        session = self._get_session(session)
        microversion = self._get_microversion(session, action='create')
        url = utils.urljoin(self.base_path, self.id, 'action')
        response = session.post(url, json=body, microversion=microversion)
        exceptions.raise_from_response(response)
        return response

    def delete(self, session, *, delete_volumes=False):
        """Delete a group."""
        body = {'delete': {'delete-volumes': delete_volumes}}
        self._action(session, body)

    def reset(self, session, status):
        """Resets the status for a group."""
        body = {'reset_status': {'status': status}}
        self._action(session, body)

    @classmethod
    def create_from_source(
        cls,
        session,
        group_snapshot_id,
        source_group_id,
        name=None,
        description=None,
    ):
        """Creates a new group from source."""
        session = cls._get_session(session)
        microversion = cls._get_microversion(session, action='create')
        url = utils.urljoin(cls.base_path, 'action')
        body = {
            'create-from-src': {
                'name': name,
                'description': description,
                'group_snapshot_id': group_snapshot_id,
                'source_group_id': source_group_id,
            }
        }
        response = session.post(url, json=body, microversion=microversion)
        exceptions.raise_from_response(response)

        group = Group()
        group._translate_response(response=response)
        return group



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
test_coe_clusters
----------------------------------

Functional tests for `shade` COE clusters methods.
"""

from openstack.tests.functional import base


class TestCompute(base.BaseFunctionalTest):
    # NOTE(flwang): Currently, running Magnum on a cloud which doesn't support
    # nested virtualization will lead to timeout. So this test file is mostly
    # like a note to document why we can't have function testing for Magnum
    # clusters CRUD.
    pass



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack import resource


class ServerUsage(resource.Resource):
    resource_key = None
    resources_key = None

    # Capabilities
    allow_create = False
    allow_fetch = False
    allow_delete = False
    allow_list = False
    allow_commit = False

    # Properties
    #: The duration that the server exists (in hours).
    hours = resource.Body('hours')
    #: The display name of a flavor.
    flavor = resource.Body('flavor')
    #: The UUID of the server.
    instance_id = resource.Body('instance_id')
    #: The server name.
    name = resource.Body('name')
    #: The UUID of the project in a multi-tenancy cloud.
    project_id = resource.Body('tenant_id')
    #: The memory size of the server (in MiB).
    memory_mb = resource.Body('memory_mb')
    #: The sum of the root disk size of the server and the ephemeral disk size
    #: of it (in GiB).
    local_gb = resource.Body('local_gb')
    #: The number of virtual CPUs that the server uses.
    vcpus = resource.Body('vcpus')
    #: The date and time when the server was launched.
    started_at = resource.Body('started_at')
    #: The date and time when the server was deleted.
    ended_at = resource.Body('ended_at')
    #: The VM state.
    state = resource.Body('state')
    #: The uptime of the server.
    uptime = resource.Body('uptime')


class Usage(resource.Resource):
    resource_key = 'tenant_usage'
    resources_key = 'tenant_usages'
    base_path = '/os-simple-tenant-usage'

    # Capabilities
    allow_create = False
    allow_fetch = True
    allow_delete = False
    allow_list = True
    allow_commit = False

    # TODO(stephenfin): Add 'start', 'end'. These conflict with the body
    # responses though.
    _query_mapping = resource.QueryParameters(
        "detailed",
        "limit",
        "marker",
        "start",
        "end",
    )

    # Properties
    #: The UUID of the project in a multi-tenancy cloud.
    project_id = resource.Body('tenant_id')
    #: A list of the server usage objects.
    server_usages = resource.Body(
        'server_usages', type=list, list_type=ServerUsage,
    )
    #: Multiplying the server disk size (in GiB) by hours the server exists,
    #: and then adding that all together for each server.
    total_local_gb_usage = resource.Body('total_local_gb_usage')
    #: Multiplying the number of virtual CPUs of the server by hours the server
    #: exists, and then adding that all together for each server.
    total_vcpus_usage = resource.Body('total_vcpus_usage')
    #: Multiplying the server memory size (in MiB) by hours the server exists,
    #: and then adding that all together for each server.
    total_memory_mb_usage = resource.Body('total_memory_mb_usage')
    #: The total duration that servers exist (in hours).
    total_hours = resource.Body('total_hours')
    #: The beginning time to calculate usage statistics on compute and storage
    #: resources.
    start = resource.Body('start')
    #: The ending time to calculate usage statistics on compute and storage
    #: resources.
    stop = resource.Body('stop')

    _max_microversion = '2.75'



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack import exceptions
from openstack import resource
from openstack import utils


class ServerMigration(resource.Resource):
    resource_key = 'migration'
    resources_key = 'migrations'
    base_path = '/servers/%(server_uuid)s/migrations'

    # capabilities
    allow_fetch = True
    allow_list = True
    allow_delete = True

    #: The ID for the server.
    server_id = resource.URI('server_uuid')

    #: The date and time when the resource was created.
    created_at = resource.Body('created_at')
    #: The target host of the migration.
    dest_host = resource.Body('dest_host')
    #: The target compute of the migration.
    dest_compute = resource.Body('dest_compute')
    #: The target node of the migration.
    dest_node = resource.Body('dest_node')
    #: The amount of disk, in bytes, that has been processed during the
    #: migration.
    disk_processed_bytes = resource.Body('disk_processed_bytes')
    #: The amount of disk, in bytes, that still needs to be migrated.
    disk_remaining_bytes = resource.Body('disk_remaining_bytes')
    #: The total amount of disk, in bytes, that needs to be migrated.
    disk_total_bytes = resource.Body('disk_total_bytes')
    #: The amount of memory, in bytes, that has been processed during the
    #: migration.
    memory_processed_bytes = resource.Body('memory_processed_bytes')
    #: The amount of memory, in bytes, that still needs to be migrated.
    memory_remaining_bytes = resource.Body('memory_remaining_bytes')
    #: The total amount of memory, in bytes, that needs to be migrated.
    memory_total_bytes = resource.Body('memory_total_bytes')
    #: The ID of the project that initiated the server migration (since
    #: microversion 2.80)
    project_id = resource.Body('project_id')
    # FIXME(stephenfin): This conflicts since there is a server ID in the URI
    # *and* in the body. We need a field that handles both or we need to use
    # different names.
    # #: The UUID of the server
    # server_id = resource.Body('server_uuid')
    #: The source compute of the migration.
    source_compute = resource.Body('source_compute')
    #: The source node of the migration.
    source_node = resource.Body('source_node')
    #: The current status of the migration.
    status = resource.Body('status')
    #: The date and time when the resource was last updated.
    updated_at = resource.Body('updated_at')
    #: The ID of the user that initiated the server migration (since
    #: microversion 2.80)
    user_id = resource.Body('user_id')
    #: The UUID of the migration (since microversion 2.59)
    uuid = resource.Body('uuid', alternate_id=True)

    _max_microversion = '2.80'

    def _action(self, session, body):
        """Preform server migration actions given the message body."""
        session = self._get_session(session)
        microversion = self._get_microversion(session, action='list')

        url = utils.urljoin(
            self.base_path % {'server_uuid': self.server_id},
            self.id,
            'action',
        )
        response = session.post(url, microversion=microversion, json=body)
        exceptions.raise_from_response(response)
        return response

    def force_complete(self, session):
        """Force on-going live migration to complete."""
        body = {'force_complete': None}
        self._action(session, body)
